@@ -192,7 +192,7 @@ void MIPSolverinstance<MIPWrapper>::processSearchAnnotations(const Annotation& a
     /// Take the variables and append them with set prioirty.
     std::vector<MIPSolverinstance::VarId> annVars;
     exprToVarArray(annotation->arg(0), annVars);
-    aPri.insert(aPri.end(), annVars.size(), --priority);
+    aPri.insert(aPri.end(), annVars.size(), static_cast<int>(--priority));
     std::move(annVars.begin(), annVars.end(), std::back_inserter(vars));
   }
 
@@ -297,13 +297,13 @@ void MIPSolverinstance<MIPWrapper>::processFlatZinc() {
   _mipWrapper->fVerbose = _options->verbose;
 
   SolveI* solveItem = getEnv()->flat()->solveItem();
+  _solveType = solveItem->st();
   VarDecl* objVd = nullptr;
 
   if (solveItem->st() != SolveI::SolveType::ST_SAT) {
     if (Id* id = solveItem->e()->dynamicCast<Id>()) {
       objVd = id->decl();
     } else {
-      std::cerr << "Objective must be Id: " << solveItem->e() << std::endl;
       throw InternalError("Objective must be Id");
     }
   }
@@ -315,8 +315,8 @@ void MIPSolverinstance<MIPWrapper>::processFlatZinc() {
     }
     VarDecl* vd = it->e();
     if (!vd->ann().isEmpty()) {
-      if (vd->ann().containsCall(constants().ann.output_array) ||
-          vd->ann().contains(constants().ann.output_var)) {
+      if (vd->ann().containsCall(Constants::constants().ann.output_array) ||
+          vd->ann().contains(Constants::constants().ann.output_var)) {
         _varsWithOutput.push_back(vd);
         //         std::cerr << (*vd);
         //         if ( vd->e() )
@@ -337,7 +337,7 @@ void MIPSolverinstance<MIPWrapper>::processFlatZinc() {
         ssm << "  VarDecl flags (ti, bt, st, ot): " << ti->type().ti() << ti->type().bt()
             << ti->type().st() << ti->type().ot() << ", dim == " << ti->type().dim()
             << "\nRemove the variable or add a constraint so it is redefined." << std::endl;
-        throw InternalError(ssm.str());
+        throw FlatteningError(getEnv()->envi(), ti->loc(), ssm.str());
       }
       double lb = 0.0;
       double ub = 1.0;  // for bool
@@ -528,27 +528,21 @@ void MIPSolverinstance<MIPWrapper>::printStatistics() {
   //   auto nn = std::chrono::system_clock::now();
   //   auto n_c = std::chrono::system_clock::to_time_t( nn );
   {
-    EnvI& env = getEnv()->envi();
-
-    std::ios oldState(nullptr);
-    oldState.copyfmt(env.outstream);
-    env.outstream.precision(12);
-    env.outstream << "%%%mzn-stat: objective=" << _mipWrapper->getObjValue() << std::endl;
-    ;
-    env.outstream << "%%%mzn-stat: objectiveBound=" << _mipWrapper->getBestBound() << std::endl;
-    ;
-    env.outstream << "%%%mzn-stat: nodes=" << _mipWrapper->getNNodes() << std::endl;
-    ;
+    auto* solns2out = getSolns2Out();
+    StatisticsStream ss(solns2out->getOutput(), solns2out->opt.flagEncapsulateJSON);
+    ss.precision(12, false);
+    if (_solveType != SolveI::SolveType::ST_SAT) {
+      ss.add("objective", _mipWrapper->getObjValue());
+      if (!std::isnan(_mipWrapper->getBestBound())) {
+        ss.add("objectiveBound", _mipWrapper->getBestBound());
+      }
+    }
+    ss.add("nodes", _mipWrapper->getNNodes());
     if (_mipWrapper->getNOpen() != 0) {
-      env.outstream << "%%%mzn-stat: openNodes=" << _mipWrapper->getNOpen() << std::endl;
+      ss.add("openNodes", _mipWrapper->getNOpen());
     };
-    env.outstream.setf(std::ios::fixed);
-    env.outstream.precision(4);
-    env.outstream << "%%%mzn-stat: solveTime=" << _mipWrapper->getWallTimeElapsed() << std::endl;
-    ;
-    env.outstream.copyfmt(oldState);
-
-    env.outstream << "%%%mzn-stat-end" << std::endl;
+    ss.precision(4, true);
+    ss.add("solveTime", _mipWrapper->getCPUTime());
   }
 }
 
@@ -697,7 +691,8 @@ inline std::string make_constraint_name(const char* pfx, int cnt,
                                         const Expression* cOrig = nullptr) {
   Call* mznp;
   std::ostringstream ss;
-  if (nullptr != cOrig && ((mznp = cOrig->ann().getCall(constants().ann.mzn_path)) != nullptr)) {
+  if (nullptr != cOrig &&
+      ((mznp = cOrig->ann().getCall(Constants::constants().ann.mzn_path)) != nullptr)) {
     assert(1 == mznp->argCount());
     auto* strp = mznp->arg(0)->dynamicCast<StringLit>();
     assert(strp);
@@ -979,8 +974,6 @@ template <class MIPWrapper>
 void p_cumulative(SolverInstanceBase& si, const Call* call) {
   auto& gi = dynamic_cast<MIPSolverinstance<MIPWrapper>&>(si);
 
-  std::unique_ptr<SECCutGen> pCG(new SECCutGen(gi.getMIPWrapper()));
-
   assert(call->argCount() == 4);
 
   std::vector<MIPSolver::Variable> startTimes;
@@ -996,6 +989,42 @@ void p_cumulative(SolverInstanceBase& si, const Call* call) {
       make_constraint_name("p_cumulative_", (gi.getMIPWrapper()->nAddedRows++), call));
 }
 
+template <class MIPWrapper>
+void p_lex_lesseq_binary(SolverInstanceBase& si, const Call* call) {
+  auto& gi = dynamic_cast<MIPSolverinstance<MIPWrapper>&>(si);
+
+  assert(call->argCount() == 3);
+
+  std::vector<MIPSolver::Variable> vec1;
+  std::vector<MIPSolver::Variable> vec2;
+  gi.exprToVarArray(call->arg(0), vec1);
+  gi.exprToVarArray(call->arg(1), vec2);
+  auto isModelCons = gi.exprToConst(call->arg(2));
+  MZN_ASSERT_HARD(vec1.size() == vec2.size());
+
+  gi.getMIPWrapper()->addLexLesseq(
+      vec1.size(), vec1.data(), vec2.data(), (bool)isModelCons,
+      make_constraint_name("p_lex_lesseq__orbisack_", (gi.getMIPWrapper()->nAddedRows++), call));
+}
+
+template <class MIPWrapper>
+void p_lex_chain_lesseq_binary(SolverInstanceBase& si, const Call* call) {
+  auto& gi = dynamic_cast<MIPSolverinstance<MIPWrapper>&>(si);
+
+  assert(call->argCount() == 5);
+
+  std::vector<MIPSolver::Variable> vars;
+  gi.exprToVarArray(call->arg(0), vars);
+  auto m = gi.exprToConst(call->arg(1));
+  auto orbitopeType = gi.exprToConst(call->arg(2));
+  auto resolveprop = gi.exprToConst(call->arg(3));
+  auto isModelCons = gi.exprToConst(call->arg(4));
+
+  gi.getMIPWrapper()->addLexChainLesseq(
+      m, vars.size() / m, vars.data(), orbitopeType, (bool)resolveprop, (bool)isModelCons,
+      make_constraint_name("p_lex_lesseq__orbisack_", (gi.getMIPWrapper()->nAddedRows++), call));
+}
+
 /// The XBZ cut generator
 template <class MIPWrapper>
 void p_xbz_cutgen(SolverInstanceBase& si, const Call* call) {
@@ -1024,7 +1053,7 @@ void p_sec_cutgen(SolverInstanceBase& si, const Call* call) {
 
   assert(call->argCount() == 1);
   gi.exprToVarArray(call->arg(0), pCG->varXij);  // WHAT ABOUT CONSTANTS?
-  const double dN = sqrt(pCG->varXij.size());
+  const double dN = sqrt(static_cast<double>(pCG->varXij.size()));
   MZN_ASSERT_HARD(fabs(dN - round(dN)) < 1e-6);  // should be a square matrix
   pCG->nN = static_cast<int>(round(dN));
   const auto sVld = pCG->validate();
@@ -1119,6 +1148,12 @@ void MIPSolverinstance<MIPWrapper>::registerConstraints() {
 
   _constraintRegistry.add("fzn_cumulative_fixed_d_r", SCIPConstraints::p_cumulative<MIPWrapper>);
 
+  _constraintRegistry.add("fzn_lex_lesseq__orbisack",
+                          SCIPConstraints::p_lex_lesseq_binary<MIPWrapper>);
+
+  _constraintRegistry.add("fzn_lex_chain_lesseq__orbitope",
+                          SCIPConstraints::p_lex_chain_lesseq_binary<MIPWrapper>);
+
   _constraintRegistry.add("bounds_disj", SCIPConstraints::p_bounds_disj<MIPWrapper>);
 
   _constraintRegistry.add("fzn_array_float_minimum", SCIPConstraints::p_array_minimum<MIPWrapper>);
