@@ -18,6 +18,7 @@
 
 #include <minizinc/flattener.hh>
 #include <minizinc/pathfileprinter.hh>
+#include <minizinc/statistics.hh>
 
 #include <fstream>
 
@@ -54,6 +55,7 @@ void Flattener::printHelp(ostream& os) const {
      << std::endl
      << "  --model-interface-only\n    Only extract parameters and output variables." << std::endl
      << "  --model-types-only\n    Only output variable (enum) type information." << std::endl
+     << "  --debug\n    Evaluate debugging intrinsics ending with _dbg." << std::endl
      << "  --no-optimize\n    Do not optimize the FlatZinc" << std::endl
      << "  --no-chain-compression\n    Do not simplify chains of implication constraints."
      << std::endl
@@ -62,6 +64,7 @@ void Flattener::printHelp(ostream& os) const {
      << std::endl
      << "  -D <data>, --cmdline-data <data>\n    Include the given data assignment in the model."
      << std::endl
+     << "  --cmdline-json-data <data>\n    Include the given JSON data in the model." << std::endl
      << "  --stdlib-dir <dir>\n    Path to MiniZinc standard library directory" << std::endl
      << "  -G <dir>, --globals-dir <dir>, --mzn-globals-dir <dir>\n    Search for included "
         "globals "
@@ -149,6 +152,10 @@ void Flattener::printHelp(ostream& os) const {
      << std::endl
      << "  --output-output-item\n    Print the output item as a string in the dzn or json output"
      << std::endl
+     << "  --only-sections <section_1,...section_n>" << std::endl
+     << "    Enable only the given comma-separated output sections." << std::endl
+     << "  --not-sections <section_1,...section_n>" << std::endl
+     << "    Disable the given comma-separated output sections." << std::endl
      << "  -Werror\n    Turn warnings into errors" << std::endl;
 }
 
@@ -174,6 +181,8 @@ bool Flattener::processOption(int& i, std::vector<std::string>& argv,
     _flags.verbose = true;
   } else if (cop.getOption("--newfzn")) {
     _flags.newfzn = true;
+  } else if (cop.getOption("--debug")) {
+    _fopts.debug = true;
   } else if (cop.getOption("--no-optimize --no-optimise")) {
     _flags.optimize = false;
   } else if (cop.getOption("--no-chain-compression")) {
@@ -216,6 +225,20 @@ bool Flattener::processOption(int& i, std::vector<std::string>& argv,
     _flags.outputObjective = true;
   } else if (cop.getOption("--output-output-item")) {
     _flags.outputOutputItem = true;
+  } else if (cop.getOption("--only-sections", &buffer)) {
+    std::stringstream ss(buffer);
+    while (ss.good()) {
+      std::string section;
+      getline(ss, section, ',');
+      _fopts.onlySections.insert(section);
+    }
+  } else if (cop.getOption("--not-sections", &buffer)) {
+    std::stringstream ss(buffer);
+    while (ss.good()) {
+      std::string section;
+      getline(ss, section, ',');
+      _fopts.notSections.insert(section);
+    }
   } else if (cop.getOption("- --input-from-stdin")) {
     _flags.stdinInput = true;
   } else if (cop.getOption("-d --data", &buffer)) {
@@ -235,6 +258,8 @@ bool Flattener::processOption(int& i, std::vector<std::string>& argv,
     // Parsed by reference
   } else if (cop.getOption("-D --cmdline-data", &buffer)) {
     _datafiles.push_back("cmd:/" + buffer);
+  } else if (cop.getOption("--cmdline-json-data", &buffer)) {
+    _datafiles.push_back("json:/" + buffer);
   } else if (cop.getOption("--allow-unbounded-vars")) {
     _flags.allowUnboundedVars = true;
   } else if (cop.getOption("--only-range-domains")) {
@@ -415,11 +440,11 @@ Env* Flattener::multiPassFlatten(const vector<unique_ptr<Pass> >& passes) {
 
   Env* pre_env = &e;
   size_t npasses = passes.size();
-  pre_env->envi().finalPassNumber = static_cast<unsigned int>(npasses);
+  pre_env->envi().multiPassInfo.finalPassNumber = static_cast<unsigned int>(npasses);
   Timer starttime;
   bool verbose = false;
   for (unsigned int i = 0; i < passes.size(); i++) {
-    pre_env->envi().currentPassNumber = i;
+    pre_env->envi().multiPassInfo.currentPassNumber = i;
     if (verbose) {
       _log << "Start pass " << i << ":\n";
     }
@@ -441,14 +466,7 @@ Env* Flattener::multiPassFlatten(const vector<unique_ptr<Pass> >& passes) {
   return pre_env;
 }
 
-class FlattenTimeout {
-public:
-  FlattenTimeout(unsigned long long int t) { GC::setTimeout(t); }
-  ~FlattenTimeout() { GC::setTimeout(0); }
-};
-
 void Flattener::flatten(const std::string& modelString, const std::string& modelName) {
-  FlattenTimeout flatten_timeout(_fopts.timeout);
   Timer flatten_time;
   _starttime.reset();
 
@@ -474,9 +492,10 @@ void Flattener::flatten(const std::string& modelString, const std::string& model
   }
 
   if (!_globalsDir.empty()) {
-    _includePaths.insert(_includePaths.begin(), _stdLibDir + "/" + _globalsDir + "/");
+    _includePaths.insert(_includePaths.begin(),
+                         FileUtils::file_path(_stdLibDir + "/" + _globalsDir + "/"));
   }
-  _includePaths.push_back(_stdLibDir + "/std/");
+  _includePaths.push_back(FileUtils::file_path(_stdLibDir + "/std/"));
 
   for (auto& includePath : _includePaths) {
     if (!FileUtils::directory_exists(includePath)) {
@@ -517,13 +536,12 @@ void Flattener::flatten(const std::string& modelString, const std::string& model
     }
   }
 
-  {
+  try {
     std::stringstream errstream;
 
     Model* m;
     _pEnv.reset(new Env(nullptr, _os, _log));
     Env* env = getEnv();
-
     if (!_flags.compileSolutionCheckModel && !_flagSolutionCheckModel.empty()) {
       // Extract variables to check from solution check model
       if (_flags.verbose) {
@@ -533,7 +551,7 @@ void Flattener::flatten(const std::string& modelString, const std::string& model
           _flagSolutionCheckModel.size() >= 4 &&
           _flagSolutionCheckModel.substr(_flagSolutionCheckModel.size() - 4) == ".mzc";
       std::vector<std::string> smm_model({_flagSolutionCheckModel});
-      Model* smm = parse(*env, smm_model, _datafiles, "", "", _includePaths, _isFlatzinc, false,
+      Model* smm = parse(*env, smm_model, _datafiles, "", "", _includePaths, {}, _isFlatzinc, false,
                          false, _flags.verbose, errstream);
       if (_flags.verbose) {
         _log << " done parsing (" << _starttime.stoptime() << ")" << std::endl;
@@ -549,24 +567,18 @@ void Flattener::flatten(const std::string& modelString, const std::string& model
         GCLock lock;
         vector<TypeError> typeErrors;
         try {
-          MiniZinc::typecheck(smm_env, smm, typeErrors, true, false, true);
+          MiniZinc::typecheck(smm_env, smm, typeErrors, true, false);
           if (!typeErrors.empty()) {
             if (!isCompressedChecker) {
-              for (auto& typeError : typeErrors) {
-                if (_flags.verbose) {
-                  _log << std::endl;
-                }
-                _log << typeError.loc() << ":" << std::endl;
-                _log << typeError.what() << ": " << typeError.msg() << std::endl;
-              }
+              throw MultipleErrors<TypeError>(typeErrors);
             }
             throw Error("multiple type errors");
           }
           for (auto& i : *smm) {
             if (auto* vdi = i->dynamicCast<VarDeclI>()) {
-              if (vdi->e()->e() == nullptr) {
+              if (vdi->e()->type().isPar() && vdi->e()->e() == nullptr) {
                 env->envi().checkVars.emplace_back(vdi->e());
-              } else if (vdi->e()->ann().contains(constants().ann.rhs_from_assignment)) {
+              } else if (vdi->e()->ann().contains(Constants::constants().ann.rhs_from_assignment)) {
                 smm_stats_oss << *vdi;
               }
             }
@@ -653,8 +665,10 @@ void Flattener::flatten(const std::string& modelString, const std::string& model
       _log << " ..." << std::endl;
     }
     errstream.str("");
+    std::unordered_set<std::string> globalInc(global_includes(_stdLibDir));
     m = parse(*env, _filenames, _datafiles, modelText, modelName.empty() ? "stdin" : modelName,
-              _includePaths, _isFlatzinc, false, false, _flags.verbose, errstream);
+              _includePaths, std::move(globalInc), _isFlatzinc, false, false, _flags.verbose,
+              errstream);
     if (!_globalsDir.empty()) {
       _includePaths.erase(_includePaths.begin());
     }
@@ -673,6 +687,7 @@ void Flattener::flatten(const std::string& modelString, const std::string& model
         std::ostringstream compiledSolutionCheckModel;
         if (_flags.compileSolutionCheckModel) {
           Printer p(compiledSolutionCheckModel, 0);
+          type_demonomorphise_library(*env, m);
           p.print(m);
         }
         GCLock lock;
@@ -682,14 +697,7 @@ void Flattener::flatten(const std::string& modelString, const std::string& model
             _flags.modelTypesOnly || _flags.modelInterfaceOnly || _flags.modelCheckOnly,
             _flags.allowMultiAssign);
         if (!typeErrors.empty()) {
-          for (auto& typeError : typeErrors) {
-            if (_flags.verbose) {
-              _log << std::endl;
-            }
-            _log << typeError.loc() << ":" << std::endl;
-            _log << typeError.what() << ": " << typeError.msg() << std::endl;
-          }
-          throw Error("multiple type errors");
+          throw MultipleErrors<TypeError>(typeErrors);
         }
         if (_flags.modelInterfaceOnly) {
           MiniZinc::output_model_interface(*env, m, _os, _includePaths);
@@ -709,6 +717,10 @@ void Flattener::flatten(const std::string& modelString, const std::string& model
           mzc_f.close();
         }
         status = SolverInstance::NONE;
+        // Not flattening so have to print warnings ourselves
+        env->dumpWarnings(_flags.encapsulateJSON ? _os : _log, _flags.werror,
+                          _flags.encapsulateJSON);
+        env->clearWarnings();
       } else {
         if (_isFlatzinc) {
           GCLock lock;
@@ -717,18 +729,11 @@ void Flattener::flatten(const std::string& modelString, const std::string& model
                               _flags.modelCheckOnly || _flags.modelInterfaceOnly,
                               _flags.allowMultiAssign, true);
           if (!typeErrors.empty()) {
-            for (auto& typeError : typeErrors) {
-              if (_flags.verbose) {
-                _log << std::endl;
-              }
-              _log << typeError.loc() << ":" << std::endl;
-              _log << typeError.what() << ": " << typeError.msg() << std::endl;
-            }
-            throw Error("multiple type errors");
+            throw MultipleErrors<TypeError>(typeErrors);
           }
           MiniZinc::register_builtins(*env);
           env->swap();
-          populate_output(*env);
+          populate_output(*env, _flags.encapsulateJSON);
         } else {
           if (_flags.verbose) {
             _log << "Flattening ...";
@@ -740,6 +745,7 @@ void Flattener::flatten(const std::string& modelString, const std::string& model
           _fopts.outputObjective = _flags.outputObjective;
           _fopts.outputOutputItem = _flags.outputOutputItem;
           _fopts.hasChecker = !_flagSolutionCheckModel.empty();
+          _fopts.encapsulateJSON = _flags.encapsulateJSON;
 #ifdef HAS_GECODE
           GecodeOptions gopts;
           gopts.onlyRangeDomains = _flags.onlyRangeDomains;
@@ -796,64 +802,67 @@ void Flattener::flatten(const std::string& modelString, const std::string& model
 
         if (_flags.statistics) {
           FlatModelStatistics stats = statistics(*env);
-          _os << "% Generated FlatZinc statistics:\n";
+          StatisticsStream ss(_os, _flags.encapsulateJSON);
 
-          _os << "%%%mzn-stat: paths=" << env->envi().getPathMap().size() << endl;
+          if (!_flags.encapsulateJSON) {
+            _os << "% Generated FlatZinc statistics:\n";
+          }
+
+          ss.add("paths", env->envi().varPathStore.getPathMap().size());
 
           if (stats.n_bool_vars != 0) {
-            _os << "%%%mzn-stat: flatBoolVars=" << stats.n_bool_vars << endl;
+            ss.add("flatBoolVars", stats.n_bool_vars);
           }
           if (stats.n_int_vars != 0) {
-            _os << "%%%mzn-stat: flatIntVars=" << stats.n_int_vars << endl;
+            ss.add("flatIntVars", stats.n_int_vars);
           }
           if (stats.n_float_vars != 0) {
-            _os << "%%%mzn-stat: flatFloatVars=" << stats.n_float_vars << endl;
+            ss.add("flatFloatVars", stats.n_float_vars);
           }
           if (stats.n_set_vars != 0) {
-            _os << "%%%mzn-stat: flatSetVars=" << stats.n_set_vars << endl;
+            ss.add("flatSetVars", stats.n_set_vars);
           }
 
           if (stats.n_bool_ct != 0) {
-            _os << "%%%mzn-stat: flatBoolConstraints=" << stats.n_bool_ct << endl;
+            ss.add("flatBoolConstraints", stats.n_bool_ct);
           }
           if (stats.n_int_ct != 0) {
-            _os << "%%%mzn-stat: flatIntConstraints=" << stats.n_int_ct << endl;
+            ss.add("flatIntConstraints", stats.n_int_ct);
           }
           if (stats.n_float_ct != 0) {
-            _os << "%%%mzn-stat: flatFloatConstraints=" << stats.n_float_ct << endl;
+            ss.add("flatFloatConstraints", stats.n_float_ct);
           }
           if (stats.n_set_ct != 0) {
-            _os << "%%%mzn-stat: flatSetConstraints=" << stats.n_set_ct << endl;
+            ss.add("flatSetConstraints", stats.n_set_ct);
           }
 
           if (stats.n_reif_ct != 0) {
-            _os << "%%%mzn-stat: evaluatedReifiedConstraints=" << stats.n_reif_ct << endl;
+            ss.add("evaluatedReifiedConstraints", stats.n_reif_ct);
           }
           if (stats.n_imp_ct != 0) {
-            _os << "%%%mzn-stat: evaluatedHalfReifiedConstraints=" << stats.n_imp_ct << endl;
+            ss.add("evaluatedHalfReifiedConstraints", stats.n_imp_ct);
           }
 
           if (stats.n_imp_del != 0) {
-            _os << "%%%mzn-stat: eliminatedImplications=" << stats.n_imp_del << endl;
+            ss.add("eliminatedImplications", stats.n_imp_del);
           }
           if (stats.n_lin_del != 0) {
-            _os << "%%%mzn-stat: eliminatedLinearConstraints=" << stats.n_lin_del << endl;
+            ss.add("eliminatedLinearConstraints", stats.n_lin_del);
           }
 
           /// Objective / SAT. These messages are used by mzn-test.py.
           SolveI* solveItem = env->flat()->solveItem();
           if (solveItem->st() != SolveI::SolveType::ST_SAT) {
             if (solveItem->st() == SolveI::SolveType::ST_MAX) {
-              _os << "%%%mzn-stat: method=\"maximize\"" << endl;
+              ss.add("method", "maximize");
             } else {
-              _os << "%%%mzn-stat: method=\"minimize\"" << endl;
+              ss.add("method", "minimize");
             }
           } else {
-            _os << "%%%mzn-stat: method=\"satisfy\"" << endl;
+            ss.add("method", "satisfy");
           }
 
-          _os << "%%%mzn-stat: flatTime=" << flatten_time.s() << endl;
-          _os << "%%%mzn-stat-end" << endl << endl;
+          ss.add("flatTime", flatten_time.s());
         }
 
         if (_flags.outputPathsStdout) {
@@ -861,7 +870,13 @@ void Flattener::flatten(const std::string& modelString, const std::string& model
             _log << "Printing Paths to stdout ..." << std::endl;
           }
           PathFilePrinter pfp(_os, env->envi());
-          pfp.print(env->flat());
+          if (_flags.encapsulateJSON) {
+            _os << "{\"type\": \"paths\", \"paths\": ";
+            pfp.json(env->flat());
+            _os << "}" << std::endl;
+          } else {
+            pfp.print(env->flat());
+          }
           if (_flags.verbose) {
             _log << " done (" << _starttime.stoptime() << ")" << std::endl;
           }
@@ -883,7 +898,9 @@ void Flattener::flatten(const std::string& modelString, const std::string& model
         if ((_fopts.collectMznPaths || _flags.twoPass) && !_flags.keepMznPaths) {
           class RemovePathAnnotations : public ItemVisitor {
           public:
-            static void removePath(Annotation& a) { a.removeCall(constants().ann.mzn_path); }
+            static void removePath(Annotation& a) {
+              a.removeCall(Constants::constants().ann.mzn_path);
+            }
             static void vVarDeclI(VarDeclI* vdi) { removePath(vdi->e()->ann()); }
             static void vConstraintI(ConstraintI* ci) { removePath(ci->e()->ann()); }
             static void vSolveI(SolveI* si) {
@@ -925,7 +942,13 @@ void Flattener::flatten(const std::string& modelString, const std::string& model
               _log << "Printing .ozn to stdout ..." << std::endl;
             }
             Printer p(_os, 0);
-            p.print(env->output());
+            Model* ozn;
+            {
+              GCLock lock;
+              ozn = copy(env->envi(), env->output());
+            }
+            type_demonomorphise_library(*env, ozn);
+            p.print(ozn);
             if (_flags.verbose) {
               _log << " done (" << _starttime.stoptime() << ")" << std::endl;
             }
@@ -936,7 +959,13 @@ void Flattener::flatten(const std::string& modelString, const std::string& model
             std::ofstream ofs(FILE_PATH(_flagOutputOzn), std::ios::out);
             check_io_status(ofs.good(), " I/O error: cannot open ozn output file. ");
             Printer p(ofs, 0);
-            p.print(env->output());
+            Model* ozn;
+            {
+              GCLock lock;
+              ozn = copy(env->envi(), env->output());
+            }
+            type_demonomorphise_library(*env, ozn);
+            p.print(ozn);
             check_io_status(ofs.good(), " I/O error: cannot write ozn output file. ");
             ofs.close();
             if (_flags.verbose) {
@@ -949,6 +978,22 @@ void Flattener::flatten(const std::string& modelString, const std::string& model
       Printer p(_os);
       p.print(m);
     }
+  } catch (ResultUndefinedError& e) {
+    // Ensure warnings are printed, but remove warning corresponding to this error
+    if (getEnv() != nullptr) {
+      getEnv()->dumpWarnings(_fopts.encapsulateJSON ? _os : _log, _flags.werror,
+                             _fopts.encapsulateJSON, e.warningIdx());
+      getEnv()->clearWarnings();
+    }
+    throw;
+  } catch (...) {
+    // Ensure warnings are printed
+    if (getEnv() != nullptr) {
+      getEnv()->dumpWarnings(_fopts.encapsulateJSON ? _os : _log, _flags.werror,
+                             _fopts.encapsulateJSON);
+      getEnv()->clearWarnings();
+    }
+    throw;
   }
 
   if (getEnv()->envi().failed()) {
@@ -957,12 +1002,14 @@ void Flattener::flatten(const std::string& modelString, const std::string& model
 
   if (_flags.verbose) {
     size_t mem = GC::maxMem();
-    if (mem < 1024) {
+    size_t kb = 1024;
+    size_t mb = kb * kb;
+    if (mem < kb) {
       _log << "Maximum memory " << mem << " bytes";
-    } else if (mem < 1024 * 1024) {
-      _log << "Maximum memory " << mem / 1024 << " Kbytes";
+    } else if (mem < mb) {
+      _log << "Maximum memory " << mem / kb << " Kbytes";
     } else {
-      _log << "Maximum memory " << mem / (1024 * 1024) << " Mbytes";
+      _log << "Maximum memory " << mem / mb << " Mbytes";
     }
     _log << "." << std::endl;
   }
