@@ -132,7 +132,7 @@ protected:
     }
     auto* newPage = static_cast<HeapPage*>(::malloc(sizeof(HeapPage) + s - 1));
     if (newPage == nullptr) {
-      throw InternalError("out of memory");
+      throw Error("out of memory");
     }
 #ifndef NDEBUG
     memset(newPage, 255, sizeof(HeapPage) + s - 1);
@@ -225,8 +225,9 @@ protected:
     // otherwise (i.e., we have been able to increase the free list, and it is now more
     // than 50% of overall allocated memory), keep threshold at allocated memory
     if ((old_free != 0 && static_cast<double>(old_free) / static_cast<double>(_freeMem) > 0.9) ||
-        static_cast<double>(_freeMem) / _allocedMem < 0.5) {
-      _gcThreshold = std::max(_min_gcThreshold, static_cast<size_t>(_allocedMem * 1.5));
+        static_cast<double>(_freeMem) / static_cast<double>(_allocedMem) < 0.5) {
+      _gcThreshold =
+          std::max(_min_gcThreshold, static_cast<size_t>(static_cast<double>(_allocedMem) * 1.5));
     } else {
       _gcThreshold = std::max(_min_gcThreshold, _allocedMem);
     }
@@ -337,29 +338,10 @@ const char* GC::Heap::_nodeid[] = {
 };
 #endif
 
-void GC::setTimeout(unsigned long long int t) {
-  if (gc() == nullptr) {
-    gc() = new GC();
-  }
-  gc()->_timeout = t;
-  gc()->_timeoutTimer.reset();
-}
-
 void GC::lock() {
   if (gc() == nullptr) {
     gc() = new GC();
   }
-  // If a timeout has been specified, first check counter
-  // before checking timer (counter is much cheaper, introduces
-  // less overhead)
-  if (gc()->_timeout > 0 && gc()->_timeoutCount++ > 500) {
-    gc()->_timeoutCount = 0;
-    if (gc()->_timeoutTimer.ms() > gc()->_timeout) {
-      gc()->_timeout = 0;
-      gc()->_timeoutCount = 0;
-      throw Timeout();
-    }
-  }
   if (gc()->_lockCount == 0) {
     gc()->_heap->rungc();
   }
@@ -384,17 +366,19 @@ const size_t GC::Heap::_fl_size[GC::Heap::_max_fl + 1] = {
     sizeof(Item) + 5 * sizeof(void*), sizeof(Item) + 6 * sizeof(void*),
 };
 
-GC::GC() : _heap(new Heap()), _lockCount(0), _timeout(0), _timeoutCount(0) {}
+GC::GC() : _heap(new Heap()), _lockCount(0) {}
 
 void GC::add(GCMarker* m) {
-  GC* gc = GC::gc();
-  if (gc->_heap->_rootset != nullptr) {
-    m->_rootsNext = gc->_heap->_rootset;
+  if (gc() == nullptr) {
+    gc() = new GC();
+  }
+  if (gc()->_heap->_rootset != nullptr) {
+    m->_rootsNext = gc()->_heap->_rootset;
     m->_rootsPrev = m->_rootsNext->_rootsPrev;
     m->_rootsPrev->_rootsNext = m;
     m->_rootsNext->_rootsPrev = m;
   } else {
-    gc->_heap->_rootset = m->_rootsNext = m->_rootsPrev = m;
+    gc()->_heap->_rootset = m->_rootsNext = m->_rootsPrev = m;
   }
 }
 
@@ -527,6 +511,9 @@ void GC::Heap::sweep() {
           case Item::II_SOL:
             static_cast<SolveI*>(n)->ann().~Annotation();
             break;
+          case Item::II_OUT:
+            static_cast<OutputI*>(n)->ann().~Annotation();
+            break;
           case ASTNode::NID_STR:
             static_cast<ASTStringData*>(n)->destroy();
             break;
@@ -646,6 +633,13 @@ size_t GC::maxMem() {
   return gc->_heap->_maxAllocedMem;
 }
 
+#if defined(MINIZINC_GC_STATS)
+std::map<int, GCStat>& GC::stats() {
+  GC* gc = GC::gc();
+  return gc->_heap->gc_stats;
+}
+#endif
+
 void* ASTNode::operator new(size_t size) { return GC::gc()->alloc(size); }
 
 void GC::addKeepAlive(KeepAlive* e) {
