@@ -18,11 +18,13 @@
  */
 
 #include <minizinc/file_utils.hh>
+#include <minizinc/flatten_internal.hh>
 #include <minizinc/json_parser.hh>
 #include <minizinc/parser.hh>
 #include <minizinc/prettyprinter.hh>
 
 #include <fstream>
+#include <regex>
 
 using namespace std;
 
@@ -39,7 +41,7 @@ std::string get_file_contents(std::ifstream& in) {
     in.seekg(0, std::ios::end);
     contents.resize(static_cast<unsigned int>(in.tellg()));
     in.seekg(0, std::ios::beg);
-    in.read(&contents[0], contents.size());
+    in.read(&contents[0], static_cast<long>(contents.size()));
     in.close();
     if (!contents.empty() && contents[0] == '@') {
       contents = MiniZinc::FileUtils::decode_base64(contents);
@@ -53,11 +55,56 @@ std::string get_file_contents(std::ifstream& in) {
 
 namespace MiniZinc {
 
+std::string ParserState::canonicalFilename(const std::string& f) const {
+  if (FileUtils::is_absolute(f) || std::string(filename).empty()) {
+    return f;
+  }
+  for (const auto& ip : includePaths) {
+    std::string fullname = FileUtils::file_path(ip + "/" + f);
+    if (FileUtils::file_exists(fullname)) {
+      return fullname;
+    }
+  }
+  std::string parentPath = FileUtils::dir_name(filename);
+  if (parentPath.empty()) {
+    parentPath = ".";
+  }
+  std::string fullname = FileUtils::file_path(parentPath + "/" + f);
+  if (FileUtils::file_exists(fullname)) {
+    return fullname;
+  }
+  return f;
+}
+
+std::unordered_set<std::string> global_includes(const std::string& stdlib) {
+  GCLock lock;
+  // Check if globals.mzn file can be found
+  if (!FileUtils::file_exists(stdlib + "/std/globals.mzn")) {
+    // Otherwise act as if there are no bad files
+    return {};
+  }
+
+  // Read globals file
+  std::ifstream ifs(FileUtils::file_path(stdlib + "/std/globals.mzn"));
+  std::string content((std::istreambuf_iterator<char>(ifs)), (std::istreambuf_iterator<char>()));
+
+  // Regular expression to find include items (assumes no include statements in comments)
+  std::regex include_item("include[[:space:]]+\"([^\"]+)\"", std::regex_constants::egrep);
+
+  // Collect all files directly included in globals.mzn
+  std::unordered_set<std::string> files;
+  for (auto inc = std::sregex_token_iterator(content.begin(), content.end(), include_item, 1);
+       inc != std::sregex_token_iterator(); ++inc) {
+    files.emplace(inc->str());
+  }
+  return files;
+}
+
 void parse(Env& env, Model*& model, const vector<string>& filenames,
            const vector<string>& datafiles, const std::string& modelString,
-           const std::string& modelStringName, const vector<string>& ip, bool isFlatZinc,
-           bool ignoreStdlib, bool parseDocComments, bool verbose, ostream& err,
-           std::vector<SyntaxError>& syntaxErrors) {
+           const std::string& modelStringName, const vector<string>& ip,
+           std::unordered_set<std::string> globalInc, bool isFlatZinc, bool ignoreStdlib,
+           bool parseDocComments, bool verbose, ostream& err) {
   vector<string> includePaths;
   for (const auto& i : ip) {
     includePaths.push_back(i);
@@ -70,28 +117,21 @@ void parse(Env& env, Model*& model, const vector<string>& filenames,
 
   if (!filenames.empty()) {
     GCLock lock;
-    model->setFilename(FileUtils::base_name(filenames[0]));
-    if (FileUtils::is_absolute(filenames[0])) {
-      files.emplace_back(model, nullptr, "", filenames[0]);
-    } else {
-      files.emplace_back(model, nullptr, "", workingDir + "/" + filenames[0]);
-    }
+    auto rootFileName = FileUtils::file_path(filenames[0], workingDir);
+    model->setFilename(rootFileName);
+    files.emplace_back(model, nullptr, "", rootFileName);
 
     for (unsigned int i = 1; i < filenames.size(); i++) {
       GCLock lock;
-      string fullName = filenames[i];
-      string baseName = FileUtils::base_name(filenames[i]);
-      if (!FileUtils::is_absolute(fullName)) {
-        fullName = workingDir + "/" + fullName;
-      }
-      bool isFzn = (baseName.compare(baseName.length() - 4, 4, ".fzn") == 0);
+      auto fullName = FileUtils::file_path(filenames[i], workingDir);
+      bool isFzn = (fullName.compare(fullName.length() - 4, 4, ".fzn") == 0);
       if (isFzn) {
         files.emplace_back(model, nullptr, "", fullName);
       } else {
         auto* includedModel = new Model;
-        includedModel->setFilename(baseName);
+        includedModel->setFilename(fullName);
         files.emplace_back(includedModel, nullptr, "", fullName);
-        seenModels.insert(pair<string, Model*>(baseName, includedModel));
+        seenModels.insert(pair<string, Model*>(fullName, includedModel));
         Location loc(ASTString(filenames[i]), 0, 0, 0, 0);
         auto* inc = new IncludeI(loc, includedModel->filename());
         inc->m(includedModel, true);
@@ -117,11 +157,19 @@ void parse(Env& env, Model*& model, const vector<string>& filenames,
   auto include_file = [&](const std::string& libname, bool builtin) {
     GCLock lock;
     auto* lib = new Model;
-    lib->setFilename(libname);
-    files.emplace_back(lib, nullptr, "./", libname, builtin);
-    seenModels.insert(pair<string, Model*>(libname, lib));
+    std::string fullname;
+    for (const auto& ip : includePaths) {
+      std::string n = FileUtils::file_path(ip + "/" + libname);
+      if (FileUtils::file_exists(n)) {
+        fullname = n;
+        break;
+      }
+    }
+    lib->setFilename(fullname);
+    files.emplace_back(lib, nullptr, "./", fullname, builtin);
+    seenModels.insert(pair<string, Model*>(fullname, lib));
     Location libloc(ASTString(model->filename()), 0, 0, 0, 0);
-    auto* libinc = new IncludeI(libloc, lib->filename());
+    auto* libinc = new IncludeI(libloc, libname);
     libinc->m(lib, true);
     model->addItem(libinc);
   };
@@ -149,86 +197,81 @@ void parse(Env& env, Model*& model, const vector<string>& filenames,
 
     std::string s;
     std::string fullname;
+    std::string basename;
     bool isFzn;
     if (!isModelString) {
       for (Model* p = m->parent(); p != nullptr; p = p->parent()) {
         if (p->filename() == f) {
-          err << "Error: cyclic includes: " << std::endl;
+          std::vector<ASTString> cycle;
           for (Model* pe = m; pe != nullptr; pe = pe->parent()) {
-            err << "  " << pe->filename() << std::endl;
+            cycle.push_back(pe->filename());
           }
-          goto error;
+          throw CyclicIncludeError(cycle);
         }
       }
       ifstream file;
-      if (FileUtils::is_absolute(f) || parentPath.empty()) {
+      if (FileUtils::is_absolute(f)) {
         fullname = f;
+        basename = FileUtils::base_name(fullname);
         if (FileUtils::file_exists(fullname)) {
           file.open(FILE_PATH(fullname), std::ios::binary);
         }
-      } else {
-        includePaths.push_back(parentPath);
-        unsigned int i = 0;
-        for (; i < includePaths.size(); i++) {
-          fullname = includePaths[i] + "/" + f;
-          if (FileUtils::file_exists(fullname)) {
-            file.open(FILE_PATH(fullname), std::ios::binary);
-            if (file.is_open()) {
-              break;
-            }
-          }
+      }
+      if (file.is_open() &&
+          FileUtils::file_path(FileUtils::dir_name(fullname)) != FileUtils::file_path(workingDir) &&
+          FileUtils::file_exists(workingDir + "/" + basename)) {
+        std::ostringstream w;
+        w << "file \"" << basename
+          << "\" included from library, but also exists in current working directory.";
+        env.envi().addWarning(w.str());
+      } else if (file.is_open() && globalInc.find(basename) != globalInc.end() &&
+                 fullname.find(includePaths.back()) == std::string::npos) {
+        std::ostringstream w;
+        w << "included file \"" << basename
+          << "\" overrides a global constraint file from the standard library. This is "
+             "deprecated. For a solver-specific redefinition of a global constraint, override "
+             "\"fzn_<global>.mzn\" instead.";
+        env.envi().addWarning(w.str());
+      }
+      for (const auto& includePath : includePaths) {
+        std::string deprecatedName = includePath + "/" + basename + ".deprecated.mzn";
+        if (FileUtils::file_exists(deprecatedName)) {
+          string deprecatedFullPath = FileUtils::file_path(deprecatedName);
+          string deprecatedBaseName = FileUtils::base_name(deprecatedFullPath);
+          string deprecatedDirName = FileUtils::dir_name(deprecatedFullPath);
+          auto* includedModel = new Model;
+          includedModel->setFilename(deprecatedName);
+          files.emplace_back(includedModel, nullptr, "", deprecatedName, isSTDLib, false);
+          seenModels.insert(pair<string, Model*>(deprecatedName, includedModel));
+          Location loc(ASTString(deprecatedName), 0, 0, 0, 0);
+          auto* inc = new IncludeI(loc, includedModel->filename());
+          inc->m(includedModel, true);
+          m->addItem(inc);
+          files.emplace_back(includedModel, inc, deprecatedDirName, deprecatedFullPath, isSTDLib,
+                             false);
         }
-        if (file.is_open() && i < includePaths.size() - 1 && parentPath == workingDir &&
-            FileUtils::file_path(includePaths[i], workingDir) != FileUtils::file_path(workingDir) &&
-            FileUtils::file_exists(workingDir + "/" + f)) {
-          err << "Warning: file " << f
-              << " included from library, but also exists in current working directory" << endl;
-        }
-        for (; i < includePaths.size(); i++) {
-          std::string deprecatedName = includePaths[i] + "/" + f + ".deprecated.mzn";
-          if (FileUtils::file_exists(deprecatedName)) {
-            string deprecatedBaseName = FileUtils::base_name(deprecatedName);
-            auto* includedModel = new Model;
-            includedModel->setFilename(deprecatedBaseName);
-            files.emplace_back(includedModel, nullptr, "", deprecatedName, isSTDLib, false);
-            seenModels.insert(pair<string, Model*>(deprecatedBaseName, includedModel));
-            Location loc(ASTString(deprecatedName), 0, 0, 0, 0);
-            auto* inc = new IncludeI(loc, includedModel->filename());
-            inc->m(includedModel, true);
-            m->addItem(inc);
-            files.emplace_back(includedModel, inc, deprecatedName, deprecatedBaseName, isSTDLib,
-                               false);
-          }
-        }
-        includePaths.pop_back();
       }
       if (!file.is_open()) {
         if (np_ii != nullptr) {
-          err << np_ii->loc().toString() << ":\n";
-          err << "MiniZinc: error in include item, cannot open file '" << f << "'." << endl;
-        } else {
-          err << "Error: cannot open file '" << f << "'." << endl;
+          throw IncludeError(env.envi(), np_ii->loc(), "Cannot open file '" + f + "'.");
         }
-        goto error;
+        throw Error("Cannot open file '" + f + "'.");
       }
       if (verbose) {
         std::cerr << "processing file '" << fullname << "'" << endl;
       }
       s = get_file_contents(file);
 
-      if (m->filepath().size() == 0) {
+      if (m->filepath().empty()) {
         m->setFilepath(fullname);
       }
       isFzn = (fullname.compare(fullname.length() - 4, 4, ".fzn") == 0);
-      isFzn |= (fullname.compare(fullname.length() - 4, 4, ".ozn") == 0);
-      isFzn |= (fullname.compare(fullname.length() - 4, 4, ".szn") == 0);
-      isFzn |= (fullname.compare(fullname.length() - 4, 4, ".mzc") == 0);
     } else {
       isFzn = false;
       fullname = f;
       s = parentPath;
     }
-    ParserState pp(fullname, s, err, files, seenModels, m, false, isFzn, isSTDLib,
+    ParserState pp(fullname, s, err, includePaths, files, seenModels, m, false, isFzn, isSTDLib,
                    parseDocComments);
     mzn_yylex_init(&pp.yyscanner);
     mzn_yyset_extra(&pp, pp.yyscanner);
@@ -237,10 +280,7 @@ void parse(Env& env, Model*& model, const vector<string>& filenames,
       mzn_yylex_destroy(pp.yyscanner);
     }
     if (pp.hadError) {
-      for (const auto& syntaxError : pp.syntaxErrors) {
-        syntaxErrors.push_back(syntaxError);
-      }
-      goto error;
+      throw MultipleErrors<SyntaxError>(pp.syntaxErrors);
     }
   }
 
@@ -249,6 +289,9 @@ void parse(Env& env, Model*& model, const vector<string>& filenames,
     if (f.size() >= 6 && f.substr(f.size() - 5, string::npos) == ".json") {
       JSONParser jp(env.envi());
       jp.parse(model, f, true);
+    } else if (f.size() >= 6 && f.substr(0, 6) == "json:/") {
+      JSONParser jp(env.envi());
+      jp.parseFromString(model, f.substr(6), true);
     } else {
       string s;
       if (f.size() > 5 && f.substr(0, 5) == "cmd:/") {
@@ -256,8 +299,7 @@ void parse(Env& env, Model*& model, const vector<string>& filenames,
       } else {
         std::ifstream file(FILE_PATH(f), std::ios::binary);
         if (!FileUtils::file_exists(f) || !file.is_open()) {
-          err << "Error: cannot open data file '" << f << "'." << endl;
-          goto error;
+          throw Error("Cannot open data file '" + f + "'.");
         }
         if (verbose) {
           std::cerr << "processing data file '" << f << "'" << endl;
@@ -265,7 +307,8 @@ void parse(Env& env, Model*& model, const vector<string>& filenames,
         s = get_file_contents(file);
       }
 
-      ParserState pp(f, s, err, files, seenModels, model, true, false, false, parseDocComments);
+      ParserState pp(f, s, err, includePaths, files, seenModels, model, true, false, false,
+                     parseDocComments);
       mzn_yylex_init(&pp.yyscanner);
       mzn_yyset_extra(&pp, pp.yyscanner);
       mzn_yyparse(&pp);
@@ -273,27 +316,19 @@ void parse(Env& env, Model*& model, const vector<string>& filenames,
         mzn_yylex_destroy(pp.yyscanner);
       }
       if (pp.hadError) {
-        for (const auto& syntaxError : pp.syntaxErrors) {
-          syntaxErrors.push_back(syntaxError);
-        }
-        goto error;
+        throw MultipleErrors<SyntaxError>(pp.syntaxErrors);
       }
     }
   }
-
-  return;
-error:
-  delete model;
-  model = nullptr;
 }
 
 Model* parse(Env& env, const vector<string>& filenames, const vector<string>& datafiles,
              const string& textModel, const string& textModelName,
-             const vector<string>& includePaths, bool isFlatZinc, bool ignoreStdlib,
-             bool parseDocComments, bool verbose, ostream& err) {
+             const vector<string>& includePaths, std::unordered_set<std::string> globalInc,
+             bool isFlatZinc, bool ignoreStdlib, bool parseDocComments, bool verbose,
+             ostream& err) {
   if (filenames.empty() && textModel.empty()) {
-    err << "Error: no model given" << std::endl;
-    return nullptr;
+    throw Error("No model given.");
   }
 
   Model* model;
@@ -301,9 +336,13 @@ Model* parse(Env& env, const vector<string>& filenames, const vector<string>& da
     GCLock lock;
     model = new Model();
   }
-  std::vector<SyntaxError> se;
-  parse(env, model, filenames, datafiles, textModel, textModelName, includePaths, isFlatZinc,
-        ignoreStdlib, parseDocComments, verbose, err, se);
+  try {
+    parse(env, model, filenames, datafiles, textModel, textModelName, includePaths,
+          std::move(globalInc), isFlatZinc, ignoreStdlib, parseDocComments, verbose, err);
+  } catch (Exception& e) {
+    delete model;
+    throw;
+  }
   return model;
 }
 
@@ -311,16 +350,14 @@ Model* parse_data(Env& env, Model* model, const vector<string>& datafiles,
                   const vector<string>& includePaths, bool isFlatZinc, bool ignoreStdlib,
                   bool parseDocComments, bool verbose, ostream& err) {
   vector<string> filenames;
-  std::vector<SyntaxError> se;
-  parse(env, model, filenames, datafiles, "", "", includePaths, isFlatZinc, ignoreStdlib,
-        parseDocComments, verbose, err, se);
+  parse(env, model, filenames, datafiles, "", "", includePaths, {}, isFlatZinc, ignoreStdlib,
+        parseDocComments, verbose, err);
   return model;
 }
 
 Model* parse_from_string(Env& env, const string& text, const string& filename,
                          const vector<string>& includePaths, bool isFlatZinc, bool ignoreStdlib,
-                         bool parseDocComments, bool verbose, ostream& err,
-                         std::vector<SyntaxError>& syntaxErrors) {
+                         bool parseDocComments, bool verbose, ostream& err) {
   vector<string> filenames;
   vector<string> datafiles;
   Model* model;
@@ -328,8 +365,13 @@ Model* parse_from_string(Env& env, const string& text, const string& filename,
     GCLock lock;
     model = new Model();
   }
-  parse(env, model, filenames, datafiles, text, filename, includePaths, isFlatZinc, ignoreStdlib,
-        parseDocComments, verbose, err, syntaxErrors);
+  try {
+    parse(env, model, filenames, datafiles, text, filename, includePaths, {}, isFlatZinc,
+          ignoreStdlib, parseDocComments, verbose, err);
+  } catch (Exception& e) {
+    delete model;
+    throw;
+  }
   return model;
 }
 
