@@ -112,6 +112,8 @@ bool Solns2Out::processOption(int& i, std::vector<std::string>& argv,
   } else if (cop.getOption("--output-raw",
                            &opt.flagOutputRaw)) {  // NOLINT: Allow repeated empty if
     // Parsed by reference
+  } else if (cop.getOption("--is-checker")) {
+    opt.flagIsChecker = true;
   } else if (opt.flagStandaloneSolns2Out) {
     std::string oznfile(argv[i]);
     if (oznfile.length() <= 4) {
@@ -157,7 +159,7 @@ void Solns2Out::initFromOzn(const std::string& filename) {
     _env = new Env();
     std::stringstream errstream;
     if ((_outputModel = parse(*_env, filenames, std::vector<std::string>(), "", "", _includePaths,
-                              false, false, false, false, errstream)) != nullptr) {
+                              {}, false, false, false, false, errstream)) != nullptr) {
       std::vector<TypeError> typeErrors;
       _env->model(_outputModel);
       MZN_ASSERT_HARD_MSG(_env, "solns2out: could not allocate Env");
@@ -195,9 +197,8 @@ void Solns2Out::restoreDefaults() {
 }
 
 void Solns2Out::parseAssignments(string& solution) {
-  std::vector<SyntaxError> se;
   unique_ptr<Model> sm(parse_from_string(*_env, solution, "solution received from solver",
-                                         _includePaths, false, true, false, false, _log, se));
+                                         _includePaths, false, true, false, false, _log));
   if (sm == nullptr) {
     throw Error("solns2out_base: could not parse solution");
   }
@@ -232,11 +233,54 @@ void Solns2Out::declNewOutput() {
   status = SolverInstance::SAT;
 }
 
-bool Solns2Out::evalOutput(const string& s_ExtraInfo) {
+void Solns2Out::printSolution(std::istream& sol, std::ostream& os, bool outputTime) {
+  if (opt.flagEncapsulateJSON) {
+    std::string t = opt.flagIsChecker ? "checker" : "solution";
+    os << "{\"type\": \"" << t << "\", ";
+    std::string line;
+    while (std::getline(sol, line)) {
+      // Remove line breaks from JSON object
+      os << line;
+    }
+    if (outputTime) {
+      os << ", \"time\": " << _starttime.ms().count();
+    }
+    os << "}\n";
+  } else {
+    os << sol.rdbuf();
+    os.clear();  // Writing sol.rdbuf() sets the fail bit if the solution was empty, so clear it
+    if (outputTime) {
+      os << "% time elapsed: " << _starttime.stoptime() << "\n";
+    }
+    if (!opt.solutionSeparator.empty()) {
+      os << opt.solutionSeparator << '\n';
+    }
+  }
+  if (opt.flagOutputFlush) {
+    os.flush();
+  }
+}
+
+void Solns2Out::parseStatistics(const std::string& stats, std::ostream& os) {
+  unique_ptr<Model> sm(parse_from_string(*_env, stats, "statistics received from solver",
+                                         _includePaths, false, true, false, false, _log));
+  if (sm == nullptr) {
+    throw Error("solns2out_base: could not parse statistics");
+  }
+  StatisticsStream ss(os, opt.flagEncapsulateJSON);
+  for (unsigned int i = 0; i < sm->size(); i++) {
+    if (auto* ai = (*sm)[i]->dynamicCast<AssignI>()) {
+      ss.add(ai->id().c_str(), *(ai->e()));
+    }
+  }
+}
+
+bool Solns2Out::evalOutput() {
   if (!_fNewSol2Print) {
     return true;
   }
-  ostringstream oss;
+
+  stringstream oss;
   if (!_checkerModel.empty()) {
     auto& checkerStream = _env->envi().checkerOutput;
     checkerStream.clear();
@@ -259,60 +303,38 @@ bool Solns2Out::evalOutput(const string& s_ExtraInfo) {
       checkerStream.flush();
       std::string line;
       if (std::getline(checkerStream, line)) {
-        _os << "% Solution checker report:\n";
-        _os << "% " << line << "\n";
-        while (std::getline(checkerStream, line)) {
+        if (opt.flagEncapsulateJSON) {
+          _os << line << "\n";
+        } else {
+          _os << "% Solution checker report:\n";
           _os << "% " << line << "\n";
+          while (std::getline(checkerStream, line)) {
+            _os << "% " << line << "\n";
+          }
         }
       }
     }
     ++stats.nSolns;
     if (opt.flagCanonicalize) {
-      if (_outStreamNonCanon != nullptr) {
-        if (_outStreamNonCanon->good()) {
-          (*_outStreamNonCanon) << oss.str();
-          (*_outStreamNonCanon) << comments;
-          if (!s_ExtraInfo.empty()) {
-            (*_outStreamNonCanon) << s_ExtraInfo;
-            if ('\n' != s_ExtraInfo.back()) {  /// TODO is this enough to check EOL?
-              (*_outStreamNonCanon) << '\n';
-            }
-          }
-          if (opt.flagOutputTime) {
-            (*_outStreamNonCanon) << "% time elapsed: " << _starttime.stoptime() << "\n";
-          }
-          if (!opt.solutionSeparator.empty()) {
-            (*_outStreamNonCanon) << opt.solutionSeparator << '\n';
-          }
-          if (opt.flagOutputFlush) {
-            _outStreamNonCanon->flush();
-          }
+      if (_outStreamNonCanon != nullptr && _outStreamNonCanon->good()) {
+        printSolution(oss, *_outStreamNonCanon, opt.flagOutputTime);
+      }
+      if (opt.flagOutputTime) {
+        // Print time as we get solutions
+        if (opt.flagEncapsulateJSON) {
+          getOutput() << "{\"type\": \"time\", \"time\": " << _starttime.ms().count() << "}\n";
+        } else {
+          getOutput() << "% time elapsed: " << _starttime.stoptime() << "\n";
         }
       }
     } else {
       if ((!opt.solutionComma.empty()) && stats.nSolns > 1) {
         getOutput() << opt.solutionComma << '\n';
       }
-      getOutput() << oss.str();
+
+      printSolution(oss, getOutput(), opt.flagOutputTime);
     }
   }
-  getOutput() << comments;  // print them now ????
-  comments = "";
-  if (!s_ExtraInfo.empty()) {
-    getOutput() << s_ExtraInfo;
-    if ('\n' != s_ExtraInfo.back()) {  /// TODO is this enough to check EOL?
-      getOutput() << '\n';
-    }
-  }
-  if (fNew && opt.flagOutputTime) {
-    getOutput() << "% time elapsed: " << _starttime.stoptime() << "\n";
-  }
-  if (fNew && !opt.flagCanonicalize && !opt.solutionSeparator.empty()) {
-    getOutput() << opt.solutionSeparator << '\n';
-  }
-  if (opt.flagOutputFlush) {
-    getOutput().flush();
-  }
   restoreDefaults();  // cleans data. evalOutput() should not be called again w/o assigning new
                       // data.
   return true;
@@ -328,12 +350,12 @@ void Solns2Out::checkSolution(std::ostream& oss) {
     GCLock lock;
     for (auto& i : *getModel()) {
       if (auto* vdi = i->dynamicCast<VarDeclI>()) {
-        if (vdi->e()->ann().contains(constants().ann.mzn_check_var)) {
+        if (vdi->e()->ann().contains(Constants::constants().ann.mzn_check_var)) {
           checker << vdi->e()->id()->str() << " = ";
           Expression* e = eval_par(getEnv()->envi(), vdi->e()->e());
           auto* al = e->dynamicCast<ArrayLit>();
           std::vector<Id*> enumids;
-          if (Call* cev = vdi->e()->ann().getCall(constants().ann.mzn_check_enum_var)) {
+          if (Call* cev = vdi->e()->ann().getCall(Constants::constants().ann.mzn_check_enum_var)) {
             auto* enumIdsAl = cev->arg(0)->cast<ArrayLit>();
             for (int j = 0; j < enumIdsAl->size(); j++) {
               enumids.push_back((*enumIdsAl)[j]->dynamicCast<Id>());
@@ -367,17 +389,16 @@ void Solns2Out::checkSolution(std::ostream& oss) {
     }
   }
 
-  MznSolver slv(oss, oss);
+  MznSolver slv(oss, _log, _starttime);
   slv.s2out.opt.solutionSeparator = "";
   try {
-    std::vector<std::string> args({"--solver", "org.minizinc.gecode_presolver"});
-    slv.run(args, checker.str(), "minizinc", "checker.mzc");
-  } catch (const LocationException& e) {
-    oss << e.loc() << ":" << std::endl;
-    oss << e.what() << ": " << e.msg() << std::endl;
+    slv.run(opt.checkerArgs, checker.str(), "minizinc", "checker.mzc");
   } catch (const Exception& e) {
-    std::string what = e.what();
-    oss << what << (what.empty() ? "" : ": ") << e.msg() << std::endl;
+    if (opt.flagEncapsulateJSON) {
+      e.json(std::cout);
+    } else {
+      e.print(oss);
+    }
   } catch (const exception& e) {
     oss << e.what() << std::endl;
   } catch (...) {
@@ -385,7 +406,12 @@ void Solns2Out::checkSolution(std::ostream& oss) {
   }
 
 #else
-  oss << "% solution checking not supported (need built-in Gecode)" << std::endl;
+  Warning w("solution checking not supported (need built-in Gecode)");
+  if (opt.flagEncapsulateJSON) {
+    w.json(oss, false);
+  } else {
+    w.print(oss, false);
+  }
 #endif
 }
 
@@ -398,13 +424,12 @@ void Solns2Out::checkStatistics(std::ostream& oss) {
   checker << "mzn_stats_failures = " << stats.nFails << ";\n";
   checker << "mzn_stats_solutions = " << stats.nSolns << ";\n";
   checker << "mzn_stats_nodes = " << stats.nNodes << ";\n";
-  checker << "mzn_stats_time = " << _starttime.ms() << ";\n";
+  checker << "mzn_stats_time = " << _starttime.ms().count() << ";\n";
 
-  MznSolver slv(oss, oss);
+  MznSolver slv(oss, oss, _starttime);
   slv.s2out.opt.solutionSeparator = "";
   try {
-    std::vector<std::string> args({"--solver", "org.minizinc.gecode_presolver"});
-    slv.run(args, checker.str(), "minizinc", "checker.mzc");
+    slv.run(opt.checkerArgs, checker.str(), "minizinc", "checker.mzc");
   } catch (const LocationException& e) {
     oss << e.loc() << ":" << std::endl;
     oss << e.what() << ": " << e.msg() << std::endl;
@@ -444,46 +469,77 @@ bool Solns2Out::evalOutputFinalInternal(bool /*b*/) {
     if ((!opt.solutionComma.empty()) && &sol != &*_sSolsCanon.begin()) {
       getOutput() << opt.solutionComma << '\n';
     }
-    getOutput() << sol;
-    if (!opt.solutionSeparator.empty()) {
-      getOutput() << opt.solutionSeparator << '\n';
-    }
+    std::stringstream oss;
+    oss << sol;
+    printSolution(oss, getOutput(), false);
   }
   return true;
 }
 
 bool Solns2Out::evalStatusMsg(SolverInstance::Status status) {
-  std::map<SolverInstance::Status, string> stat2msg;
-  stat2msg[SolverInstance::OPT] = opt.searchCompleteMsg;
-  stat2msg[SolverInstance::UNSAT] = opt.unsatisfiableMsg;
-  stat2msg[SolverInstance::UNBND] = opt.unboundedMsg;
-  stat2msg[SolverInstance::UNSATorUNBND] = opt.unsatorunbndMsg;
-  stat2msg[SolverInstance::UNKNOWN] = opt.unknownMsg;
-  stat2msg[SolverInstance::ERROR] = opt.errorMsg;
-  stat2msg[SolverInstance::NONE] = "";
-  auto it = stat2msg.find(status);
-  if (stat2msg.end() != it) {
-    getOutput() << comments;
-    if (!it->second.empty()) {
-      getOutput() << it->second << '\n';
+  getOutput() << comments;
+  comments = "";
+
+  Solns2Out::status = status;
+  std::string label;
+  switch (status) {
+    case SolverInstance::SAT:
+      if (opt.flagOutputFlush) {
+        getOutput().flush();
+      }
+      return true;
+    case SolverInstance::OPT: {
+      if (opt.flagEncapsulateJSON) {
+        bool sat = getEnv()->flat()->solveItem()->st() == SolveI::ST_SAT;
+        label = sat ? "ALL_SOLUTIONS" : "OPTIMAL_SOLUTION";
+      } else {
+        label = opt.searchCompleteMsg;
+      }
+      break;
+    }
+    case SolverInstance::UNSAT:
+      label = opt.flagEncapsulateJSON ? "UNSATISFIABLE" : opt.unsatisfiableMsg;
+      break;
+    case SolverInstance::UNBND:
+      label = opt.flagEncapsulateJSON ? "UNBOUNDED" : opt.unboundedMsg;
+      break;
+    case SolverInstance::UNSATorUNBND:
+      label = opt.flagEncapsulateJSON ? "UNSAT_OR_UNBOUNDED" : opt.unsatorunbndMsg;
+      break;
+    case SolverInstance::UNKNOWN:
+      label = opt.flagEncapsulateJSON ? "UNKNOWN" : opt.unknownMsg;
+      break;
+    case SolverInstance::ERROR:
+      label = opt.flagEncapsulateJSON ? "ERROR" : opt.errorMsg;
+      break;
+    case SolverInstance::NONE:
+      label = "";
+      break;
+    default:
+      assert(false);
+  }
+  if (opt.flagEncapsulateJSON) {
+    if (label.empty()) {
+      getOutput() << "{\"type\": \"time\", \"time\": " << _starttime.ms().count() << "}\n";
+    } else {
+      getOutput() << "{\"type\": \"status\", \"status\": \"" << label << "\"";
+      if (opt.flagOutputTime) {
+        getOutput() << ", \"time\": " << _starttime.ms().count();
+      }
+      getOutput() << "}\n";
+    }
+  } else {
+    if (!label.empty()) {
+      getOutput() << label << '\n';
     }
     if (opt.flagOutputTime) {
       getOutput() << "% time elapsed: " << _starttime.stoptime() << "\n";
     }
-    if (opt.flagOutputFlush) {
-      getOutput().flush();
-    }
-    Solns2Out::status = status;
-  } else {
-    getOutput() << comments;
-    if (opt.flagOutputFlush) {
-      getOutput().flush();
-    }
-    MZN_ASSERT_HARD_MSG(SolverInstance::SAT == status,  // which is ignored
-                        "solns2out_base: undefined solution status code " << status);
-    Solns2Out::status = SolverInstance::SAT;
   }
-  comments = "";
+
+  if (opt.flagOutputFlush) {
+    getOutput().flush();
+  }
   return true;
 }
 
@@ -506,7 +562,6 @@ void Solns2Out::init() {
           FileUtils::inflate_string(_statisticsCheckerModel);
         }
       } else {
-        GCLock lock;
         _declmap.insert(make_pair(vdi->e()->id()->str(), DE(vdi->e(), vdi->e()->e())));
       }
     }
@@ -583,6 +638,11 @@ bool Solns2Out::feedRawDataChunk(const char* data) {
     }
     auto it = _mapInputStatus.find(line);
     if (_mapInputStatus.end() != it) {
+      if (!_stats.empty()) {
+        // Didn't have %%%mzn-stat-end, but should print stats anyway at this point
+        parseStatistics(_stats, getOutput());
+        _stats.clear();
+      }
       if (SolverInstance::SAT == it->second) {
         parseAssignments(solution);
         evalOutput();
@@ -596,17 +656,33 @@ bool Solns2Out::feedRawDataChunk(const char* data) {
         char c = '_';
         iss >> skipws >> c;
         if (iss.good() && '%' == c) {
-          // Feed comments directly
-          getOutput() << line << '\n';
+          bool is_statistic = line.substr(0, 13) == "%%%mzn-stat: ";
+          std::ostringstream message;
+          if (opt.flagEncapsulateJSON) {
+            if (is_statistic) {
+              _stats += line.substr(13) + ";";
+            } else if (line == "%%%mzn-stat-end") {
+              parseStatistics(_stats, message);
+              _stats.clear();
+            } else {
+              message << "{\"type\": \"comment\", \"comment\": \"" << Printer::escapeStringLit(line)
+                      << "\"}\n";
+            }
+          } else {
+            message << line << '\n';
+          }
+          getOutput() << message.str();
           if (opt.flagOutputFlush) {
             getOutput().flush();
           }
-          if (_outStreamNonCanon != nullptr) {
-            if (_outStreamNonCanon->good()) {
-              (*_outStreamNonCanon) << line << '\n';
+          if (_outStreamNonCanon != nullptr && _outStreamNonCanon->good()) {
+            (*_outStreamNonCanon) << message.str();
+            if (opt.flagOutputFlush) {
+              _outStreamNonCanon->flush();
             }
           }
-          if (line.substr(0, 13) == "%%%mzn-stat: " && line.size() > 13) {
+
+          if (is_statistic && line.size() > 13) {
             if (line.substr(13, 6) == "nodes=") {
               std::istringstream iss(line.substr(19));
               int n_nodes;
@@ -639,15 +715,20 @@ void Solns2Out::createInputMap() {
   _mapInputStatus[opt.unboundedMsgDef] = SolverInstance::UNBND;
   _mapInputStatus[opt.unsatorunbndMsgDef] = SolverInstance::UNSATorUNBND;
   _mapInputStatus[opt.unknownMsgDef] = SolverInstance::UNKNOWN;
-  _mapInputStatus[opt.errorMsg] = SolverInstance::ERROR;
+  _mapInputStatus[opt.errorMsgDef] = SolverInstance::ERROR;
 }
 
 void Solns2Out::printStatistics(ostream& os) {
-  os << "%%%mzn-stat: nSolutions=" << stats.nSolns << "\n";
+  if (!_stats.empty()) {
+    // Didn't have %%%mzn-stat-end, but should print stats anyway at this point
+    parseStatistics(_stats, getOutput());
+    _stats.clear();
+  }
+  StatisticsStream ss(os, opt.flagEncapsulateJSON);
+  ss.add("nSolutions", stats.nSolns);
   if (!_statisticsCheckerModel.empty()) {
     std::ostringstream oss;
     checkStatistics(oss);
-    os << "%%%mzn-stat: statisticsCheck=\"" << Printer::escapeStringLit(oss.str()) << "\"\n";
+    ss.add("statisticsCheck", oss.str());
   }
-  os << "%%%mzn-stat-end\n";
 }
