@@ -30,16 +30,11 @@
 #include <cstdlib>
 #include <ctime>
 #include <fstream>
+#include <future>
 #include <iomanip>
 #include <iostream>
 #include <ratio>
 
-#ifdef HAS_GUROBI
-#include <minizinc/solvers/MIP/MIP_gurobi_solverfactory.hh>
-#endif
-#ifdef HAS_CPLEX
-#include <minizinc/solvers/MIP/MIP_cplex_solverfactory.hh>
-#endif
 #ifdef HAS_OSICBC
 #include <minizinc/solvers/MIP/MIP_osicbc_solverfactory.hh>
 #endif
@@ -55,6 +50,12 @@
 #ifdef HAS_SCIP
 #include <minizinc/solvers/MIP/MIP_scip_solverfactory.hh>
 #endif
+#ifdef HAS_CPLEX
+#include <minizinc/solvers/MIP/MIP_cplex_solverfactory.hh>
+#endif
+#ifdef HAS_GUROBI
+#include <minizinc/solvers/MIP/MIP_gurobi_solverfactory.hh>
+#endif
 #include <minizinc/solvers/fzn_solverfactory.hh>
 #include <minizinc/solvers/fzn_solverinstance.hh>
 #include <minizinc/solvers/mzn_solverfactory.hh>
@@ -66,12 +67,6 @@ using namespace std;
 using namespace MiniZinc;
 
 SolverInitialiser::SolverInitialiser() {
-#ifdef HAS_GUROBI
-  GurobiSolverFactoryInitialiser _gurobi_init;
-#endif
-#ifdef HAS_CPLEX
-  static CplexSolverFactoryInitialiser _cplex_init;
-#endif
 #ifdef HAS_OSICBC
   static OSICBCSolverFactoryInitialiser _osicbc_init;
 #endif
@@ -86,8 +81,14 @@ SolverInitialiser::SolverInitialiser() {
 #endif
 #ifdef HAS_SCIP
   static SCIPSolverFactoryInitialiser _scip_init;
+#endif
+#ifdef HAS_CPLEX
+  static CplexSolverFactoryInitialiser _cplex_init;
 #endif
   static FZNSolverFactoryInitialiser _fzn_init;
+#ifdef HAS_GUROBI
+  static GurobiSolverFactoryInitialiser _gurobi_init;
+#endif
   static MZNSolverFactoryInitialiser _mzn_init;
   static NLSolverFactoryInitialiser _nl_init;
 }
@@ -162,8 +163,9 @@ void SolverFactory::destroySI(SolverInstanceBase* pSI) {
   _sistorage.erase(it);
 }
 
-MznSolver::MznSolver(std::ostream& os0, std::ostream& log0)
-    : _solverConfigs(log0),
+MznSolver::MznSolver(std::ostream& os0, std::ostream& log0, const Timer& startTime)
+    : _startTime(startTime),
+      _solverConfigs(log0),
       _flt(os0, log0, _solverConfigs.mznlibDir()),
       _executableName("<executable>"),
       _os(os0),
@@ -247,7 +249,9 @@ void MznSolver::printHelp(const std::string& selectedSolver) {
       << "  --compiler-statistics\n    Print statistics for compilation." << std::endl
       << "  -c, --compile\n    Compile only (do not run solver)." << std::endl
       << "  --config-dirs\n    Output configuration directories." << std::endl
-      << "  --param-file <file>\n    Load parameters from the given JSON file." << std::endl;
+      << "  --param-file <file>\n    Load parameters from the given JSON file." << std::endl
+      << "  --json-stream\n    Print output as newline-delimited JSON message objects."
+      << std::endl;
 
   if (selectedSolver.empty()) {
     _flt.printHelp(_os);
@@ -361,47 +365,43 @@ MznSolver::OptionStatus MznSolver::processOptions(std::vector<std::string>& argv
     }
 
     if (!paramFile.empty()) {
-      try {
-        auto paramFilePath = FileUtils::file_path(paramFile, workingDirs.back());
-        if (std::find(paramFiles.begin(), paramFiles.end(), paramFilePath) != paramFiles.end()) {
-          throw ParamException("Cyclic parameter configuration file");
-        }
-        // add parameter file arguments
-        ParamConfig pc;
-        pc.blacklist(
-            {"--solvers", "--solvers-json", "--solver-json", "--help", "-h", "--config-dirs"});
-        pc.negatedFlag("-i", "-n-i");
-        pc.negatedFlag("--intermediate", "--no-intermediate");
-        pc.negatedFlag("--intermediate-solutions", "--no-intermediate-solutions");
-        pc.negatedFlag("--all-satisfaction", "--disable-all-satisfaction");
-        pc.load(paramFilePath);
+      auto paramFilePath = FileUtils::file_path(paramFile, workingDirs.back());
+      if (std::find(paramFiles.begin(), paramFiles.end(), paramFilePath) != paramFiles.end()) {
+        throw ParamException("Cyclic parameter configuration file");
+      }
+      // add parameter file arguments
+      ParamConfig pc;
+      pc.blacklist(
+          {"--solvers", "--solvers-json", "--solver-json", "--help", "-h", "--config-dirs"});
+      pc.negatedFlag("-i", "-n-i");
+      pc.negatedFlag("--intermediate", "--no-intermediate");
+      pc.negatedFlag("--intermediate-solutions", "--no-intermediate-solutions");
+      pc.negatedFlag("--all-satisfaction", "--disable-all-satisfaction");
+      pc.load(paramFilePath);
 
-        // Insert the new options
-        auto toInsert = pc.argv();
-        auto remove = argv.begin() + (usedFlag ? i - 1 : i);
-        auto position = argv.erase(remove, argv.begin() + i + 1);
-        if (pushWorkingDir) {
-          position =
-              argv.insert(position, {"--push-working-directory",
-                                     FileUtils::file_path(FileUtils::dir_name(paramFilePath))}) +
-              2;
-        }
-        position = argv.insert(position, toInsert.begin(), toInsert.end()) + toInsert.size();
-        if (pushWorkingDir) {
-          position = argv.insert(position, "--pop-working-directory") + 1;
-        }
-        paramFiles.push_back(paramFilePath);
-        argc = argv.size();
+      // Insert the new options
+      auto toInsert = pc.argv();
+      auto remove = argv.begin() + (usedFlag ? i - 1 : i);
+      auto position = argv.erase(remove, argv.begin() + i + 1);
+      if (pushWorkingDir) {
+        position =
+            argv.insert(position, {"--push-working-directory",
+                                   FileUtils::file_path(FileUtils::dir_name(paramFilePath))}) +
+            2;
+      }
+      position = argv.insert(position, toInsert.begin(), toInsert.end()) +
+                 static_cast<int>(toInsert.size());
+      if (pushWorkingDir) {
+        position = argv.insert(position, "--pop-working-directory") + 1;
+      }
+      paramFiles.push_back(paramFilePath);
+      argc = static_cast<int>(argv.size());
 
-        // Have to process the newly added options
-        if (usedFlag) {
-          i -= 2;
-        } else {
-          i--;
-        }
-      } catch (ParamException& e) {
-        _log << "Solver parameter exception: " << e.msg() << endl;
-        return OPTION_ERROR;
+      // Have to process the newly added options
+      if (usedFlag) {
+        i -= 2;
+      } else {
+        i--;
       }
     }
   }
@@ -450,7 +450,7 @@ MznSolver::OptionStatus MznSolver::processOptions(std::vector<std::string>& argv
       }
     }
     argv = remaining;
-    argc = remaining.size();
+    argc = static_cast<int>(remaining.size());
   }
   for (auto* sf : get_global_solver_registry()->getSolverFactories()) {
     // Notify solver factories that factory flags are done
@@ -461,6 +461,12 @@ MznSolver::OptionStatus MznSolver::processOptions(std::vector<std::string>& argv
   _solverConfigs.populate(_log);
 
   for (i = 1; i < argc; ++i) {
+    if (argv[i] == "--backend-flags") {
+      // Ensure that we don't consume the argument to --backend-flags
+      argv[j++] = argv[i++];
+      argv[j++] = argv[i];
+      continue;
+    }
     if (argv[i] == "-h" || argv[i] == "--help") {
       if (argc > i + 1) {
         printHelp(argv[i + 1]);
@@ -527,7 +533,15 @@ MznSolver::OptionStatus MznSolver::processOptions(std::vector<std::string>& argv
         _log << "Argument required for --time-limit" << endl;
         return OPTION_ERROR;
       }
-      flagOverallTimeLimit = atoi(argv[i].c_str());
+      flagOverallTimeLimit = std::chrono::milliseconds(atoi(argv[i].c_str()));
+    } else if (argv[i] == "-t" || argv[i] == "--solver-time-limit" ||
+               argv[i] == "--fzn-time-limit") {
+      ++i;
+      if (i == argc) {
+        _log << "Argument required for --solver-time-limit" << endl;
+        return OPTION_ERROR;
+      }
+      flagSolverTimeLimit = std::chrono::milliseconds(atoi(argv[i].c_str()));
     } else if (argv[i] == "--solver") {
       ++i;
       if (i == argc) {
@@ -549,8 +563,24 @@ MznSolver::OptionStatus MznSolver::processOptions(std::vector<std::string>& argv
     } else if (argv[i] == "-s" || argv[i] == "--statistics") {
       flagStatistics = true;
       flagCompilerStatistics = true;
+    } else if (argv[i] == "-r" || argv[i] == "--seed" || argv[i] == "--random-seed") {
+      ++i;
+      if (i == argc) {
+        _log << "Argument required for --random-seed" << endl;
+        return OPTION_ERROR;
+      }
+      flagRandomSeed = true;
+      randomSeed = atoi(argv[i].c_str());
     } else if (argv[i] == "--compiler-statistics") {
       flagCompilerStatistics = true;
+    } else if (argv[i] == "--json-stream") {
+      flagEncapsulateJSON = true;
+      s2out.opt.checkerArgs.emplace_back("--json-stream");
+    } else if (argv[i] == "--only-sections" || argv[i] == "--not-sections") {
+      argv[j++] = argv[i];
+      s2out.opt.checkerArgs.push_back(argv[i++]);
+      argv[j++] = argv[i];
+      s2out.opt.checkerArgs.push_back(argv[i]);
     } else {
       if ((argv[i] == "--fzn-cmd" || argv[i] == "--flatzinc-cmd") && solver.empty()) {
         solver = "org.minizinc.mzn-fzn";
@@ -583,6 +613,7 @@ MznSolver::OptionStatus MznSolver::processOptions(std::vector<std::string>& argv
   _flt.setFlagOutputByDefault(ifMzn2Fzn());
 
   bool isMznMzn = false;
+  s2out.opt.flagEncapsulateJSON = flagEncapsulateJSON;
 
   if (!flagIsSolns2out) {
     try {
@@ -607,6 +638,8 @@ MznSolver::OptionStatus MznSolver::processOptions(std::vector<std::string>& argv
           _supportsA = true;
         } else if (flag == "-i") {
           _supportsI = true;
+        } else if (flag == "--json-stream") {
+          _supportsJSONStream = true;
         }
       }
 
@@ -647,6 +680,9 @@ MznSolver::OptionStatus MznSolver::processOptions(std::vector<std::string>& argv
               } else {
                 additionalArgs_s.push_back(sc.executable());
               }
+              if (!sc.passFlags().empty()) {
+                add_flags("--mzn-flag", sc.passFlags(), additionalArgs_s);
+              }
 
               if (!fzn_mzn_flags.empty()) {
                 add_flags("--mzn-flag", fzn_mzn_flags, additionalArgs_s);
@@ -696,6 +732,13 @@ MznSolver::OptionStatus MznSolver::processOptions(std::vector<std::string>& argv
               } else {
                 additionalArgs.push_back(sc.executable());
               }
+              if (!sc.passFlags().empty()) {
+                if (sc.supportsFzn()) {
+                  add_flags("--fzn-flag", sc.passFlags(), additionalArgs);
+                } else {
+                  add_flags("--nl-flag", sc.passFlags(), additionalArgs);
+                }
+              }
               if (!fzn_mzn_flags.empty()) {
                 if (sc.supportsFzn()) {
                   add_flags("--fzn-flag", fzn_mzn_flags, additionalArgs);
@@ -764,7 +807,7 @@ MznSolver::OptionStatus MznSolver::processOptions(std::vector<std::string>& argv
               addedArgs.push_back(argv[i]);
             }
             argv = addedArgs;
-            argc = addedArgs.size();
+            argc = static_cast<int>(addedArgs.size());
           }
           break;
         }
@@ -838,10 +881,52 @@ MznSolver::OptionStatus MznSolver::processOptions(std::vector<std::string>& argv
 }
 
 void MznSolver::flatten(const std::string& modelString, const std::string& modelName) {
+  std::exception_ptr exc;
   _flt.setFlagVerbose(flagCompilerVerbose);
   _flt.setFlagStatistics(flagCompilerStatistics);
-  _flt.setFlagTimelimit(flagOverallTimeLimit);
-  _flt.flatten(modelString, modelName);
+  _flt.setFlagEncapsulateJSON(flagEncapsulateJSON);
+  if (flagRandomSeed) {
+    _flt.setRandomSeed(randomSeed);
+  }
+
+#ifndef __EMSCRIPTEN__
+  // Create timing thread
+  std::promise<void> done;
+  auto done_future = done.get_future();
+  std::packaged_task<int()> timer([&] {
+    auto time_left = std::max(std::chrono::milliseconds(1), flagOverallTimeLimit - _startTime.ms());
+    if (flagOverallTimeLimit != std::chrono::milliseconds(0)) {
+      auto time_left =
+          std::max(std::chrono::milliseconds(1), flagOverallTimeLimit - _startTime.ms());
+      if (done_future.wait_for(time_left) == std::future_status::timeout) {
+        // Trigger time out!
+        _flt.cancel();
+      }
+    }
+    return time_left.count();
+  });
+  auto timer_future = timer.get_future();
+  std::thread thr(std::move(timer));
+#endif
+
+  // Flattening process in main thread
+  try {
+    _flt.flatten(modelString, modelName);
+  } catch (...) {
+    // Capture exception to gracefully handle threads
+    exc = std::current_exception();
+  }
+
+#ifndef __EMSCRIPTEN__
+  // Join timing thread
+  done.set_value();
+  thr.join();
+#endif
+
+  // Rethrow exception if necessary
+  if (exc) {
+    std::rethrow_exception(exc);
+  }
 }
 
 SolverInstance::Status MznSolver::solve() {
@@ -866,8 +951,6 @@ SolverInstance::Status MznSolver::solve() {
 SolverInstance::Status MznSolver::run(const std::vector<std::string>& args0,
                                       const std::string& model, const std::string& exeName,
                                       const std::string& modelName) {
-  using namespace std::chrono;
-  steady_clock::time_point startTime = steady_clock::now();
   std::vector<std::string> args = {exeName};
   for (const auto& a : args0) {
     args.push_back(a);
@@ -896,6 +979,12 @@ SolverInstance::Status MznSolver::run(const std::vector<std::string>& args0,
   }
 
   if (!ifMzn2Fzn() && _sf->getId() == "org.minizinc.mzn-mzn") {
+    if (_supportsJSONStream && flagEncapsulateJSON) {
+      std::vector<std::string> json_stream_flag(1);
+      json_stream_flag[0] = "--json-stream";
+      int i = 0;
+      _sf->processOption(_siOpt, i, json_stream_flag);
+    }
     Env env;
     _si = _sf->createSI(env, _log, _siOpt);
     _si->setSolns2Out(&s2out);
@@ -911,26 +1000,33 @@ SolverInstance::Status MznSolver::run(const std::vector<std::string>& args0,
   try {
     flatten(model, modelName);
   } catch (Timeout&) {
+    if (ifMzn2Fzn()) {
+      throw;
+    }
     s2out.evalStatus(SolverInstance::UNKNOWN);
     return SolverInstance::UNKNOWN;
   }
 
-  if (!ifMzn2Fzn() && flagOverallTimeLimit != 0) {
-    steady_clock::time_point afterFlattening = steady_clock::now();
-    milliseconds passed = duration_cast<milliseconds>(afterFlattening - startTime);
-    milliseconds time_limit(flagOverallTimeLimit);
-    if (passed > time_limit) {
-      s2out.evalStatus(getFltStatus());
-      return SolverInstance::UNKNOWN;
+  if (!ifMzn2Fzn()) {
+    if (flagOverallTimeLimit + flagSolverTimeLimit > std::chrono::milliseconds(0)) {
+      std::chrono::milliseconds time_left(0);
+      if (flagOverallTimeLimit == std::chrono::milliseconds(0)) {
+        time_left = flagSolverTimeLimit;
+      } else if (flagSolverTimeLimit == std::chrono::milliseconds(0)) {
+        time_left = flagOverallTimeLimit - _startTime.ms();
+      } else {
+        time_left = std::min(flagSolverTimeLimit, flagOverallTimeLimit - _startTime.ms());
+      }
+      std::vector<std::string> timeoutArgs(
+          {"--solver-time-limit", std::to_string(time_left.count())});
+      int i = 0;
+      _sf->processOption(_siOpt, i, timeoutArgs);
+    }
+    if (flagRandomSeed) {
+      std::vector<std::string> randomArgs({"--random-seed", std::to_string(randomSeed)});
+      int i = 0;
+      _sf->processOption(_siOpt, i, randomArgs);
     }
-    int time_left = (time_limit - passed).count();
-    std::vector<std::string> timeoutArgs(2);
-    timeoutArgs[0] = "--solver-time-limit";
-    std::ostringstream oss;
-    oss << time_left;
-    timeoutArgs[1] = oss.str();
-    int i = 0;
-    _sf->processOption(_siOpt, i, timeoutArgs);
   }
 
   if (SolverInstance::UNKNOWN == getFltStatus()) {
