@@ -27,12 +27,12 @@
 #include <string>
 
 #ifdef CPLEX_PLUGIN
-#ifdef HAS_DLFCN_H
-#include <dlfcn.h>
-#elif defined HAS_WINDOWS_H
+#ifdef _WIN32
 #define NOMINMAX  // Ensure the words min/max remain available
 #include <Windows.h>
 #undef ERROR
+#else
+#include <dlfcn.h>
 #endif
 #endif
 
@@ -44,7 +44,12 @@ using namespace std;
 
 namespace {
 void* dll_open(const std::string& file) {
-#ifdef HAS_DLFCN_H
+#ifdef _WIN32
+  if (MiniZinc::FileUtils::is_absolute(file)) {
+    return LoadLibrary(file.c_str());
+  }
+  return LoadLibrary((file + ".dll").c_str());
+#else
   if (MiniZinc::FileUtils::is_absolute(file)) {
     return dlopen(file.c_str(), RTLD_NOW);
   }
@@ -52,29 +57,24 @@ void* dll_open(const std::string& file) {
     return so;
   }
   return dlopen(("lib" + file + ".jnilib").c_str(), RTLD_NOW);
-#else
-  if (MiniZinc::FileUtils::is_absolute(file)) {
-    return LoadLibrary(file.c_str());
-  }
-  return LoadLibrary((file + ".dll").c_str());
 #endif
 }
 void* dll_sym(void* dll, const char* sym) {
-#ifdef HAS_DLFCN_H
-  void* ret = dlsym(dll, sym);
-#else
+#ifdef _WIN32
   void* ret = GetProcAddress((HMODULE)dll, sym);
+#else
+  void* ret = dlsym(dll, sym);
 #endif
   if (ret == nullptr) {
-    throw MiniZinc::InternalError("cannot load symbol " + string(sym) + " from CPLEX dll");
+    throw MiniZinc::Error("cannot load symbol " + string(sym) + " from CPLEX dll");
   }
   return ret;
 }
 void dll_close(void* dll) {
-#ifdef HAS_DLFCN_H
-  dlclose(dll);
-#else
+#ifdef _WIN32
   FreeLibrary((HMODULE)dll);
+#else
+  dlclose(dll);
 #endif
 }
 }  // namespace
@@ -102,9 +102,9 @@ void MIPCplexWrapper::checkDLL() {
 
   if (_cplexDll == nullptr) {
     if (_factoryOptions.cplexDll.empty()) {
-      throw MiniZinc::InternalError("cannot load cplex dll, specify --cplex-dll");
+      throw MiniZinc::Error("cannot load cplex dll, specify --cplex-dll");
     }
-    throw MiniZinc::InternalError("cannot load cplex dll `" + _factoryOptions.cplexDll + "'");
+    throw MiniZinc::Error("cannot load cplex dll `" + _factoryOptions.cplexDll + "'");
   }
 
   *(void**)(&dll_CPXaddfuncdest) = dll_sym(_cplexDll, "CPXaddfuncdest");
@@ -238,7 +238,7 @@ string MIPCplexWrapper::getDescription(FactoryOptions& factoryOpt,
     } else {
       v += "[?? ...cannot open CPLEX env to query version]";
     }
-  } catch (MiniZinc::InternalError&) {
+  } catch (MiniZinc::Error&) {
     v += "[?? ...cannot open CPLEX env to query version]";
   }
   v += "  Compiled  " __DATE__ "  " __TIME__;
@@ -260,7 +260,7 @@ string MIPCplexWrapper::getVersion(FactoryOptions& factoryOpt,
     } else {
       v += "<unknown version>";
     }
-  } catch (MiniZinc::InternalError&) {
+  } catch (MiniZinc::Error&) {
     v += "<unknown version>";
   }
   return v;
@@ -276,7 +276,7 @@ vector<string> MIPCplexWrapper::getRequiredFlags(FactoryOptions& f) {
     if (env != nullptr) {
       return {};
     }
-  } catch (MiniZinc::InternalError&) {
+  } catch (MiniZinc::Error&) {
   }
   return {"--cplex-dll"};
 }
@@ -506,7 +506,7 @@ vector<MiniZinc::SolverConfig::ExtraFlag> MIPCplexWrapper::getExtraFlags(
       res.emplace_back("--cplex-" + param, param, param_type, param_range, param_default);
     }
     return res;
-  } catch (MiniZinc::InternalError&) {
+  } catch (MiniZinc::Error&) {
     return {};
   }
   return {};
@@ -604,11 +604,10 @@ void MIPCplexWrapper::wrapAssert(bool cond, const string& msg, bool fTerm) {
     strcpy(_cplexBuffer, "[NO ERROR STRING GIVEN]");
     dll_CPXgeterrorstring(_env, _status, _cplexBuffer);
     string msgAll = ("  MIPCplexWrapper runtime error:  " + msg + "  " + _cplexBuffer);
-    cerr << msgAll << endl;
     if (fTerm) {
-      cerr << "TERMINATING." << endl;
-      throw runtime_error(msgAll);
+      throw MiniZinc::Error(msgAll);
     }
+    cerr << msgAll << endl;
   }
 }
 
@@ -675,10 +674,10 @@ void MIPCplexWrapper::doAddVars(size_t n, double* obj, double* lb, double* ub,
         ctype[i] = CPX_BINARY;
         break;
       default:
-        throw runtime_error("  MIPWrapper: unknown variable type");
+        throw MiniZinc::InternalError("  MIPWrapper: unknown variable type");
     }
   }
-  _status = dll_CPXnewcols(_env, _lp, n, obj, lb, ub, &ctype[0], &pcNames[0]);
+  _status = dll_CPXnewcols(_env, _lp, static_cast<int>(n), obj, lb, ub, &ctype[0], &pcNames[0]);
   wrapAssert(_status == 0, "Failed to declare variables.");
 }
 
@@ -691,7 +690,7 @@ static char get_cplex_constr_cense(MIPWrapper::LinConType sense) {
     case MIPWrapper::GQ:
       return 'G';
     default:
-      throw runtime_error("  MIPCplexWrapper: unknown constraint type");
+      throw MiniZinc::InternalError("  MIPCplexWrapper: unknown constraint type");
   }
 }
 
@@ -741,14 +740,14 @@ bool MIPCplexWrapper::addWarmStart(const std::vector<VarId>& vars,
   if (_status != 0) {  // not existent
     // status = dll_CPXaddmipstarts (env, lp, mcnt, nzcnt, beg, varindices,
     //                            values, effortlevel, mipstartname);
-    _status = dll_CPXaddmipstarts(_env, _lp, 1, vars.size(), &beg, vars.data(), vals.data(),
-                                  nullptr, (char**)&sMSName);
+    _status = dll_CPXaddmipstarts(_env, _lp, 1, static_cast<int>(vars.size()), &beg, vars.data(),
+                                  vals.data(), nullptr, (char**)&sMSName);
     wrapAssert(_status == 0, "Failed to add warm start.");
   } else {
     // status = dll_CPXchgmipstarts (env, lp, mcnt, mipstartindices, nzcnt, beg, varindices, values,
     // effortlevel);
-    _status = dll_CPXchgmipstarts(_env, _lp, 1, &msindex, vars.size(), &beg, vars.data(),
-                                  vals.data(), nullptr);
+    _status = dll_CPXchgmipstarts(_env, _lp, 1, &msindex, static_cast<int>(vars.size()), &beg,
+                                  vars.data(), vals.data(), nullptr);
     wrapAssert(_status == 0, "Failed to extend warm start.");
   }
   return true;
@@ -1002,20 +1001,20 @@ static int CPXPUBLIC myusercutcallback(CPXCENVptr env, void* cbdata, int wherefr
     }
     MIPWrapper::CutInput cutInput;
     info->cutcbfn(outpRlx, cutInput, info->psi, fMIPSol);
-    static int nCuts = 0;
-    nCuts += cutInput.size();
+    // static int nCuts = 0;
+    // nCuts += cutInput.size();
     // if ( cutInput.size() )
     //  cerr << "\n   N CUTS:  " << nCuts << endl;
     for (auto& cd : cutInput) {
       if ((cd.mask & (MIPWrapper::MaskConsType_Usercut | MIPWrapper::MaskConsType_Lazy)) == 0) {
-        throw runtime_error("Cut callback: should be user/lazy");
+        throw MiniZinc::InternalError("Cut callback: should be user/lazy");
       }
       /* Use a cut violation tolerance of 0.01 */
       if (true) {  // NOLINT: cutvio > 0.01 ) {
-        status = cw->dll_CPXcutcallbackadd(env, cbdata, wherefrom, cd.rmatind.size(), cd.rhs,
-                                           get_cplex_constr_cense(cd.sense), cd.rmatind.data(),
-                                           cd.rmatval.data(),
-                                           CPX_USECUT_FORCE);  // PURGE?
+        status = cw->dll_CPXcutcallbackadd(
+            env, cbdata, wherefrom, static_cast<int>(cd.rmatind.size()), cd.rhs,
+            get_cplex_constr_cense(cd.sense), cd.rmatind.data(), cd.rmatval.data(),
+            CPX_USECUT_FORCE);  // PURGE?
         if (status != 0) {
           fprintf(stderr, "CPLEX callback: failed to add cut.\n");
           goto TERMINATE;
@@ -1070,7 +1069,7 @@ MIPCplexWrapper::Status MIPCplexWrapper::convertStatus(int cplexStatus) {
       break;
       //      case CPXMIP_ABORT_INFEAS:
     case CPXMIP_FAIL_INFEAS:
-      s = Status::__ERROR;
+      s = Status::ERROR_STATUS;
       break;
     default:
       //      case CPXMIP_OPTIMAL_TOL:
@@ -1137,7 +1136,8 @@ void MIPCplexWrapper::solve() {  // Move into ancestor?
     wrapAssert(_status == 0, "Failed to set CPXPARAM_TimeLimit.", false);
   }
   if (_options->nSolLimit > 0) {
-    _status = dll_CPXsetintparam(_env, CPXPARAM_MIP_Limits_Solutions, _options->nSolLimit);
+    _status = dll_CPXsetintparam(_env, CPXPARAM_MIP_Limits_Solutions,
+                                 static_cast<CPXINT>(_options->nSolLimit));
     wrapAssert(_status == 0, "Failed to set CPXPARAM_MIP_Limits_Solutions.", false);
   }
   if (_options->nSeed >= 0) {
@@ -1179,7 +1179,7 @@ void MIPCplexWrapper::solve() {  // Move into ancestor?
   //    wrapAssert(!status, "Failed to set CPXPARAM_MIP_Tolerances_ObjDifference.", false);
 
   /// Solution callback
-  output.nCols = colObj.size();
+  output.nCols = static_cast<int>(colObj.size());
   _x.resize(output.nCols);
   output.x = &_x[0];
   if (_options->flagIntermediate && (cbui.solcbfn != nullptr)) {
@@ -1292,10 +1292,11 @@ void MIPCplexWrapper::solve() {  // Move into ancestor?
 
   output.dWallTime =
       std::chrono::duration<double>(std::chrono::steady_clock::now() - output.dWallTime0).count();
-  double tmNow = std::clock();
+  auto tmNow = std::clock();
   // _status = dll_CPXgettime (_env, &tmNow);   Buggy in 12.7.1.0
   wrapAssert(_status == 0, "Failed to get time stamp.", false);
-  output.dCPUTime = (tmNow - cbui.pOutput->cCPUTime0) / CLOCKS_PER_SEC;
+  output.dCPUTime =
+      static_cast<double>(tmNow - cbui.pOutput->cCPUTime0) / static_cast<double>(CLOCKS_PER_SEC);
 
   int solstat = dll_CPXgetstat(_env, _lp);
   output.status = convertStatus(solstat);
