@@ -36,7 +36,10 @@ ScipPlugin::ScipPlugin()
     : Plugin(
 #ifdef _WIN32
           {
-            "libscip", "scip", "C:\\Program Files\\SCIPOptSuite 7.0.1\\bin\\libscip.dll",
+            "libscip", "scip", "C:\\Program Files\\SCIPOptSuite 8.0.0\\bin\\libscip.dll",
+                "C:\\Program Files\\SCIPOptSuite 7.0.3\\bin\\libscip.dll",
+                "C:\\Program Files\\SCIPOptSuite 7.0.2\\bin\\libscip.dll",
+                "C:\\Program Files\\SCIPOptSuite 7.0.1\\bin\\libscip.dll",
                 "C:\\Program Files\\SCIPOptSuite 7.0.0\\bin\\libscip.dll",
                 "C:\\Program Files\\SCIPOptSuite 6.0.2\\bin\\scip.dll",
                 "C:\\Program Files\\SCIPOptSuite 6.0.1\\bin\\scip.dll",
@@ -88,6 +91,8 @@ void ScipPlugin::load() {
   load_symbol(SCIPcreateConsBasicIndicator);
   load_symbol(SCIPcreateConsBasicBounddisjunction);
   load_symbol(SCIPcreateConsBasicCumulative);
+  load_symbol(SCIPcreateConsBasicOrbisack);
+  load_symbol(SCIPcreateConsBasicOrbitope);
   load_symbol(SCIPgetNSolsFound);
   load_symbol(SCIPgetNSols);
   load_symbol(SCIPsetIntParam);
@@ -249,14 +254,14 @@ bool MIPScipWrapper::Options::processOption(int& i, vector<string>& argv,
   } else if (string(argv[i]) == "-f") {  // NOLINT: Allow repeated empty if
     //     std::cerr << "  Flag -f: ignoring fixed strategy anyway." << std::endl;
   } else if (cop.get("--writeModel", &buffer)) {
-    sExportModel = MiniZinc::FileUtils::file_path(buffer);
+    sExportModel = MiniZinc::FileUtils::file_path(buffer, workingDir);
   } else if (cop.get("-p --parallel", &nThreads)) {        // NOLINT: Allow repeated empty if
   } else if (cop.get("--solver-time-limit", &nTimeout)) {  // NOLINT: Allow repeated empty if
   } else if (cop.get("--workmem", &nWorkMemLimit)) {       // NOLINT: Allow repeated empty if
   } else if (cop.get("--readParam", &buffer)) {
-    sReadParams = MiniZinc::FileUtils::file_path(buffer);
+    sReadParams = MiniZinc::FileUtils::file_path(buffer, workingDir);
   } else if (cop.get("--writeParam", &buffer)) {
-    sWriteParams = MiniZinc::FileUtils::file_path(buffer);
+    sWriteParams = MiniZinc::FileUtils::file_path(buffer, workingDir);
   } else if (cop.get("--absGap", &absGap)) {  // NOLINT: Allow repeated empty if
   } else if (cop.get("--relGap", &relGap)) {  // NOLINT: Allow repeated empty if
   } else if (cop.get("--intTol", &intTol)) {  // NOLINT: Allow repeated empty if
@@ -427,8 +432,8 @@ SCIP_RETCODE MIPScipWrapper::addRowSCIP(int nnz, int* rmatind, double* rmatval,
                                         MIPWrapper::LinConType sense, double rhs, int mask,
                                         const string& rowName) {
   /// Convert var types:
-  double lh = -SCIPinfinityPlugin(_plugin, _scip);
-  double rh = SCIPinfinityPlugin(_plugin, _scip);
+  double lh = -getInfBound();
+  double rh = getInfBound();
   switch (sense) {
     case LQ:
       rh = rhs;
@@ -539,7 +544,7 @@ void MIPScipWrapper::addBoundsDisj(int n, double* fUB, double* bnd, int* vars, i
   }
 
   SCIP_PLUGIN_CALL(_plugin->SCIPcreateConsBasicBounddisjunction(
-      _scip, &cons, rowName.c_str(), v.size(), v.data(), bt.data(), bs.data()));
+      _scip, &cons, rowName.c_str(), static_cast<int>(v.size()), v.data(), bt.data(), bs.data()));
   SCIP_PLUGIN_CALL(_plugin->SCIPaddCons(_scip, cons));
   SCIP_PLUGIN_CALL(_plugin->SCIPreleaseCons(_scip, &cons));
 }
@@ -564,6 +569,48 @@ void MIPScipWrapper::addCumulative(int nnz, int* rmatind, double* d, double* r,
   SCIP_PLUGIN_CALL(_plugin->SCIPreleaseCons(_scip, &cons));
 }
 
+/// Lex-lesseq binary, currently SCIP only
+/// TODO check all variables are binary, SCIP 7.0.2 does not
+void MIPScipWrapper::addLexLesseq(int nnz, int* rmatind1, int* rmatind2, bool isModelCons,
+                                  const std::string& rowName) {
+  SCIP_CONS* cons;
+  vector<SCIP_VAR*> vars1(nnz);
+  vector<SCIP_VAR*> vars2(nnz);
+
+  for (int j = 0; j < nnz; ++j) {
+    vars1[j] = _scipVars[rmatind1[j]];
+    vars2[j] = _scipVars[rmatind2[j]];
+  }
+
+  SCIP_PLUGIN_CALL(_plugin->SCIPcreateConsBasicOrbisack(
+      _scip, &cons, rowName.c_str(), vars2.data(), vars1.data(),  // it's actually lex_greatereq
+      nnz, FALSE, FALSE, (SCIP_Bool)isModelCons));
+  SCIP_PLUGIN_CALL(_plugin->SCIPaddCons(_scip, cons));
+  SCIP_PLUGIN_CALL(_plugin->SCIPreleaseCons(_scip, &cons));
+}
+
+/// Lex-chain-lesseq binary, currently SCIP only
+void MIPScipWrapper::addLexChainLesseq(int m, int n, int* rmatind, int nOrbitopeType,
+                                       bool resolveprop, bool isModelCons,
+                                       const std::string& rowName) {
+  SCIP_CONS* cons;
+  vector<vector<SCIP_VAR*> > vars(m, vector<SCIP_VAR*>(size_t(n)));
+  vector<SCIP_VAR**> vars_data(m);
+
+  for (int i = 0; i < m; ++i) {
+    for (int j = 0; j < n; ++j) {
+      vars[i][j] = _scipVars[rmatind[i * n + (n - j - 1)]];  // it's actually lex_chain_greatereq
+    }
+    vars_data[i] = vars[i].data();
+  }
+
+  SCIP_PLUGIN_CALL(_plugin->SCIPcreateConsBasicOrbitope(
+      _scip, &cons, rowName.c_str(), vars_data.data(), (SCIP_ORBITOPETYPE)nOrbitopeType, m, n,
+      (SCIP_Bool)resolveprop, (SCIP_Bool)isModelCons));
+  SCIP_PLUGIN_CALL(_plugin->SCIPaddCons(_scip, cons));
+  SCIP_PLUGIN_CALL(_plugin->SCIPreleaseCons(_scip, &cons));
+}
+
 void MIPScipWrapper::addTimes(int x, int y, int z, const string& rowName) {
   /// As x*y - z == 0
   double zCoef = -1.0;
@@ -583,8 +630,13 @@ void MIPScipWrapper::addTimes(int x, int y, int z, const string& rowName) {
 #define EVENTHDLR_NAME "bestsol"
 #define EVENTHDLR_DESC "event handler for best solutions found"
 
+namespace {
 // Dirty way of accessing SCIP functions inside C callbacks
-static ScipPlugin* _cb_plugin;
+ScipPlugin* _cb_plugin;
+
+MIPWrapper::CBUserInfo* cbuiPtr = nullptr;
+SCIP_VAR** _scipVarsPtr = nullptr;
+}  // namespace
 
 /** initialization method of event handler (called after problem was transformed) */
 static SCIP_DECL_EVENTINIT(eventInitBestsol) { /*lint --e{715}*/
@@ -612,9 +664,6 @@ static SCIP_DECL_EVENTEXIT(eventExitBestsol) { /*lint --e{715}*/
   return SCIP_OKAY;
 }
 
-static MIPWrapper::CBUserInfo* cbuiPtr = nullptr;
-static SCIP_VAR** _scipVarsPtr = nullptr;
-
 /** execution method of event handler */
 static SCIP_DECL_EVENTEXEC(eventExecBestsol) { /*lint --e{715}*/
   SCIP_SOL* bestsol;
@@ -654,7 +703,11 @@ static SCIP_DECL_EVENTEXEC(eventExecBestsol) { /*lint --e{715}*/
     cbuiPtr->pOutput->nOpenNodes = _cb_plugin->SCIPgetNNodesLeft(scip);
     cbuiPtr->pOutput->bestBound = _cb_plugin->SCIPgetDualbound(scip);
 
-    cbuiPtr->pOutput->dCPUTime = -1;
+    cbuiPtr->pOutput->dWallTime = std::chrono::duration<double>(std::chrono::steady_clock::now() -
+                                                                cbuiPtr->pOutput->dWallTime0)
+                                      .count();
+    cbuiPtr->pOutput->dCPUTime =
+        double(std::clock() - cbuiPtr->pOutput->cCPUTime0) / CLOCKS_PER_SEC;
 
     /// Call the user function:
     if (cbuiPtr->solcbfn != nullptr) {
@@ -780,12 +833,13 @@ SCIP_RETCODE MIPScipWrapper::solveSCIP() {  // Move into ancestor?
   }
 
   //     assert(_scipVars.size() == colObj.size());
-  int cur_numcols = _scipVars.size();  // No, we create negated indicators: getNCols();
+  int cur_numcols =
+      static_cast<int>(_scipVars.size());  // No, we create negated indicators: getNCols();
   assert(cur_numcols == colObj.size());
   assert(cur_numcols == _scipVars.size());
 
   /// Solution callback
-  output.nCols = colObj.size();
+  output.nCols = static_cast<int>(colObj.size());
   _x.resize(output.nCols);
   output.x = &_x[0];
   if (_options->flagIntermediate && cbui.solcbfn != nullptr && cbuiPtr == nullptr) {
@@ -843,7 +897,7 @@ SCIP_RETCODE MIPScipWrapper::solveSCIP() {  // Move into ancestor?
   }
 
   cbui.pOutput->dWallTime0 = output.dWallTime0 = std::chrono::steady_clock::now();
-  output.dCPUTime = clock();
+  output.dCPUTime = static_cast<double>(clock());
 
   /* Optimize the problem and obtain solution. */
   SCIP_PLUGIN_CALL_R(_plugin, _plugin->SCIPsolve(_scip));
@@ -851,7 +905,7 @@ SCIP_RETCODE MIPScipWrapper::solveSCIP() {  // Move into ancestor?
 
   output.dWallTime =
       std::chrono::duration<double>(std::chrono::steady_clock::now() - output.dWallTime0).count();
-  output.dCPUTime = (clock() - output.dCPUTime) / CLOCKS_PER_SEC;
+  output.dCPUTime = (static_cast<double>(clock()) - output.dCPUTime) / CLOCKS_PER_SEC;
 
   cbuiPtr = nullptr;  /// cleanup
   _scipVarsPtr = nullptr;
